from __future__ import annotations

import asyncio
import functools
import logging
import uuid
from collections.abc import (
    Awaitable,
    Callable,
    Iterable,
    Mapping,
    MutableMapping,
    Sequence,
)
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Concatenate,
    Final,
    NamedTuple,
    Protocol,
    TypeVar,
    cast,
)
from uuid import UUID

import graphene
import graphql
import sqlalchemy as sa
from aiodataloader import DataLoader
from aiotools import apartial
from graphene.types import Scalar
from graphene.types.scalars import MAX_INT, MIN_INT
from graphene_federation import shareable
from graphql import GraphQLError, Undefined
from graphql.language.ast import FloatValueNode, IntValueNode, ObjectValueNode, ValueNode
from graphql.language.printer import print_ast
from sqlalchemy.engine.result import Result
from sqlalchemy.engine.row import Row
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.orm.attributes import InstrumentedAttribute

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.errors.api import InvalidAPIParameters
from ai.backend.manager.errors.common import GenericForbidden, ObjectNotFound
from ai.backend.manager.models.minilang.ordering import (
    OrderDirection,
    OrderingItem,
    QueryOrderParser,
)
from ai.backend.manager.models.minilang.queryfilter import QueryFilterParser, WhereClauseType
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.models.utils import execute_with_retry

from .gql_relay import (
    AsyncListConnectionField,
    AsyncNode,
    ConnectionPaginationOrder,
)

if TYPE_CHECKING:
    from sqlalchemy.orm.attributes import InstrumentedAttribute
    from sqlalchemy.sql.selectable import ScalarSelect

    from .schema import GraphQueryContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

SAFE_MIN_INT = -9007199254740991
SAFE_MAX_INT = 9007199254740991

DEFAULT_PAGE_SIZE: Final[int] = 10


@shareable
class ResourceLimit(graphene.ObjectType):  # type: ignore[misc]
    key = graphene.String()
    min = graphene.String()
    max = graphene.String(
        deprecation_reason="Deprecated since 25.14.0. The max slot limit validation has been removed as it was deemed obsolete."
    )


@shareable
class KVPair(graphene.ObjectType):  # type: ignore[misc]
    key = graphene.String()
    value = graphene.String()


class ResourceLimitInput(graphene.InputObjectType):  # type: ignore[misc]
    key = graphene.String()
    min = graphene.String()
    max = graphene.String(
        deprecation_reason="Deprecated since 25.14.0. The max slot limit validation has been removed as it was deemed obsolete."
    )


class KVPairInput(graphene.InputObjectType):  # type: ignore[misc]
    key = graphene.String()
    value = graphene.String()


class BigInt(Scalar):  # type: ignore[misc]
    """
    BigInt is an extension of the regular graphene.Int scalar type
    to support integers outside the range of a signed 32-bit integer.
    """

    @staticmethod
    def coerce_bigint(value: int | str) -> int | float:
        num = int(value)
        if not (SAFE_MIN_INT <= num <= SAFE_MAX_INT):
            raise ValueError("Cannot serialize integer out of the safe range.")
        if not (MIN_INT <= num <= MAX_INT):
            # treat as float
            return float(int(num))
        return num

    serialize = coerce_bigint
    parse_value = coerce_bigint

    @staticmethod
    def parse_literal(node: graphql.language.ast.ValueNode) -> int | float | None:
        if isinstance(node, graphql.language.ast.IntValueNode):
            num = int(node.value)
            if not (SAFE_MIN_INT <= num <= SAFE_MAX_INT):
                raise ValueError("Cannot parse integer out of the safe range.")
            if not (MIN_INT <= num <= MAX_INT):
                # treat as float
                return float(int(num))
            return num
        return None


class Bytes(Scalar):  # type: ignore[misc]
    class Meta:
        description = "Added in 24.09.1."

    @staticmethod
    def serialize(val: bytes) -> str:
        return val.hex()

    @staticmethod
    def parse_literal(node: Any, _variables: dict[str, Any] | None = None) -> bytes | None:
        if isinstance(node, graphql.language.ast.StringValueNode):
            return bytes.fromhex(node.value)
        return None

    @staticmethod
    def parse_value(value: str) -> bytes:
        return bytes.fromhex(value)


class ImageRefType(graphene.InputObjectType):  # type: ignore[misc]
    name = graphene.String(required=True)
    registry = graphene.String()
    architecture = graphene.String()


class UUIDFloatMap(Scalar):  # type: ignore[misc]
    """
    Added in 25.4.0.
    Verifies that the key is a UUID (represented as a string) and the value is a float.
    """

    @staticmethod
    def serialize(value: Any) -> dict[str, float]:
        if not isinstance(value, dict):
            raise GraphQLError(f"UUIDFloatMap cannot represent non-dict value: {value!r}")

        validated: dict[str, float] = {}
        for k, v in value.items():
            try:
                key_str = str(uuid.UUID(k))
            except ValueError as e:
                raise GraphQLError(f"UUIDFloatMap cannot represent key {k} as a valid UUID") from e

            if not isinstance(v, float):
                raise GraphQLError(f"UUIDFloatMap cannot represent value {v} as a float")
            validated[key_str] = v
        return validated

    @classmethod
    def parse_literal(
        cls, node: ValueNode, _variables: dict[str, Any] | None = None
    ) -> dict[str, float]:
        if not isinstance(node, ObjectValueNode):
            raise GraphQLError(f"UUIDFloatMap cannot represent non-object value: {print_ast(node)}")
        validated: dict[str, Any] = {}
        for field in node.fields:
            key = field.name.value
            if isinstance(field.value, (FloatValueNode, IntValueNode)):
                try:
                    validated[key] = float(field.value.value)
                except Exception as e:
                    raise GraphQLError(
                        f"UUIDFloatMap cannot represent value for key {key} as a float"
                    ) from e
            else:
                raise GraphQLError(
                    f"UUIDFloatMap cannot represent non-numeric value for key {key}: {print_ast(field.value)}"
                )
        return validated

    @staticmethod
    def parse_value(value: Any) -> dict[str, float]:
        if not isinstance(value, dict):
            raise GraphQLError(f"UUIDFloatMap cannot represent non-dict value: {value!r}")
        validated: dict[str, float] = {}
        for k, v in value.items():
            try:
                key_str = str(uuid.UUID(k))
            except ValueError as e:
                raise GraphQLError(f"UUIDFloatMap cannot represent key {k} as a valid UUID") from e
            if not isinstance(v, float):
                raise GraphQLError(f"UUIDFloatMap cannot represent value {v} as a float")
            validated[key_str] = v
        return validated


def extract_object_uuid(info: graphene.ResolveInfo, global_id: str, object_name: str) -> UUID:
    """
    Converts a GraphQL global ID to its corresponding UUID.
    If the global ID is not valid, raises an error using the provided object name.
    """

    _, raw_id = AsyncNode.resolve_global_id(info, global_id)
    if not raw_id:
        raw_id = global_id

    try:
        return UUID(raw_id)
    except ValueError as e:
        raise ObjectNotFound(object_name) from e


# DataLoader-related types and classes
TContext = TypeVar("TContext")
TLoaderKey = TypeVar("TLoaderKey")
TLoaderResult = TypeVar("TLoaderResult")


def _build_gql_type_cache() -> dict[str, Any]:
    """Build a cache mapping type names to their classes from gql_legacy submodules."""
    import importlib
    from pathlib import Path

    cache: dict[str, Any] = {}
    gql_legacy_path = Path(__file__).parent

    for py_file in gql_legacy_path.glob("*.py"):
        if py_file.name.startswith("_"):
            continue
        module_name = py_file.stem
        try:
            submod = importlib.import_module(f".{module_name}", "ai.backend.manager.api.gql_legacy")
            for attr_name in getattr(submod, "__all__", dir(submod)):
                if not attr_name.startswith("_"):
                    attr = getattr(submod, attr_name, None)
                    if isinstance(attr, type):
                        cache[attr_name] = attr
        except ImportError:
            continue
    return cache


class DataLoaderManager[TContext, TLoaderKey, TLoaderResult]:
    """
    For every different combination of filtering conditions, we need to make a
    new DataLoader instance because it "batches" the database queries.
    This manager get-or-creates dataloaders with fixed conditions (represetned
    as arguments) like a cache.

    NOTE: Just like DataLoaders, it is recommended to instantiate this manager
    for every incoming API request.
    """

    cache: dict[int, DataLoader[TLoaderKey, TLoaderResult]]
    _type_cache: dict[str, Any] | None = None

    def __init__(self) -> None:
        self.cache = {}

    @classmethod
    def _get_type_cache(cls) -> dict[str, Any]:
        if cls._type_cache is None:
            cls._type_cache = _build_gql_type_cache()
        return cls._type_cache

    @staticmethod
    def _get_key(otname: str, args: tuple[Any, ...], kwargs: dict[str, Any]) -> int:
        """
        Calculate the hash of the all arguments and keyword arguments.
        """
        key = (otname,) + args
        for item in kwargs.items():
            key += item
        return hash(key)

    def load_attr(self, objtype_name: str) -> Any:
        """Load a GraphQL type class by name from gql_legacy submodules."""
        type_cache = self._get_type_cache()
        if objtype_name not in type_cache:
            raise AttributeError(f"Type '{objtype_name}' not found in gql_legacy submodules")
        return type_cache[objtype_name]

    def get_loader(
        self, context: GraphQueryContext, objtype_name: str, *args: Any, **kwargs: Any
    ) -> DataLoader[Any, Any]:
        k = self._get_key(objtype_name, args, kwargs)
        loader = self.cache.get(k)
        if loader is None:
            objtype_name, has_variant, variant_name = objtype_name.partition(".")
            objtype = self.load_attr(objtype_name)
            if has_variant:
                batch_load_fn = getattr(objtype, "batch_load_" + variant_name)
            else:
                batch_load_fn = objtype.batch_load
            loader = DataLoader(
                apartial(batch_load_fn, context, *args, **kwargs),
                max_batch_size=128,
            )
            self.cache[k] = loader
        return loader

    @staticmethod
    def _get_func_key(
        func: Callable[
            Concatenate[TContext, Sequence[TLoaderKey], ...],
            Awaitable[Sequence[TLoaderResult]],
        ],
        **kwargs: Any,
    ) -> int:
        func_and_kwargs = (func, *[(k, kwargs[k]) for k in sorted(kwargs.keys())])
        return hash(func_and_kwargs)

    def get_loader_by_func(
        self,
        context: TContext,
        batch_load_func: Callable[
            Concatenate[TContext, Sequence[TLoaderKey], ...],
            Awaitable[Sequence[TLoaderResult]],
        ],
        # Using kwargs-only to prevent argument position confusion
        # when DataLoader calls `batch_load_func(keys)` which is `partial(batch_load_func, **kwargs)(keys)`.
        **kwargs: Any,
    ) -> DataLoader[TLoaderKey, TLoaderResult]:
        async def batch_load_wrapper(keys: Sequence[TLoaderKey]) -> list[TLoaderResult]:
            # aiodataloader always converts the result via list(),
            # so we can enforce type-casting here.
            return cast(list[TLoaderResult], await batch_load_func(context, keys, **kwargs))

        key = self._get_func_key(batch_load_func, **kwargs)
        loader = self.cache.get(key)
        if loader is None:
            loader = DataLoader(
                batch_load_wrapper,
                max_batch_size=128,
            )
            self.cache[key] = loader
        return loader


class Item(graphene.Interface):  # type: ignore[misc]
    id = graphene.ID()


class PaginatedList(graphene.Interface):  # type: ignore[misc]
    items = graphene.List(Item, required=True)
    total_count = graphene.Int(required=True)


# ref: https://github.com/python/mypy/issues/1212
T_SQLBasedGQLObject = TypeVar("T_SQLBasedGQLObject", bound="_SQLBasedGQLObject")
T_Key = TypeVar("T_Key")


class _SQLBasedGQLObject(Protocol):
    @classmethod
    def from_row(
        cls: type[T_SQLBasedGQLObject],
        ctx: GraphQueryContext,
        row: Row[Any] | DeclarativeMeta,
    ) -> T_SQLBasedGQLObject: ...


async def batch_result(
    graph_ctx: GraphQueryContext,
    db_conn: SAConnection | SASession,
    query: sa.sql.Select[Any],
    obj_type: type[T_SQLBasedGQLObject],
    key_list: Iterable[T_Key],
    key_getter: Callable[[Row[Any]], T_Key],
) -> Sequence[T_SQLBasedGQLObject | None]:
    """
    A batched query adaptor for (key -> item) resolving patterns.
    """
    objs_per_key: dict[T_Key, T_SQLBasedGQLObject | None]
    objs_per_key = dict()
    for key in key_list:
        objs_per_key[key] = None
    if isinstance(db_conn, SASession):
        async for row in await db_conn.stream_scalars(query):
            objs_per_key[key_getter(row)] = obj_type.from_row(graph_ctx, row)
    else:
        async for row in await db_conn.stream(query):
            objs_per_key[key_getter(row)] = obj_type.from_row(graph_ctx, row)
    return [*objs_per_key.values()]


async def batch_multiresult(
    graph_ctx: GraphQueryContext,
    db_conn: SAConnection | SASession,
    query: sa.sql.Select[Any],
    obj_type: type[T_SQLBasedGQLObject],
    key_list: Iterable[T_Key],
    key_getter: Callable[[Row[Any]], T_Key],
) -> Sequence[Sequence[T_SQLBasedGQLObject]]:
    """
    A batched query adaptor for (key -> [item]) resolving patterns.
    """
    objs_per_key: dict[T_Key, list[T_SQLBasedGQLObject]]
    objs_per_key = dict()
    for key in key_list:
        objs_per_key[key] = list()
    if isinstance(db_conn, SASession):
        async for row in await db_conn.stream_scalars(query):
            objs_per_key[key_getter(row)].append(obj_type.from_row(graph_ctx, row))
    else:
        async for row in await db_conn.stream(query):
            objs_per_key[key_getter(row)].append(obj_type.from_row(graph_ctx, row))
    return [*objs_per_key.values()]


async def batch_result_in_session(
    graph_ctx: GraphQueryContext,
    db_sess: SASession,
    query: sa.sql.Select[Any],
    obj_type: type[T_SQLBasedGQLObject],
    key_list: Iterable[T_Key],
    key_getter: Callable[[Row[Any]], T_Key],
) -> Sequence[T_SQLBasedGQLObject | None]:
    """
    A batched query adaptor for (key -> item) resolving patterns.
    stream the result in async session.
    """
    objs_per_key: dict[T_Key, T_SQLBasedGQLObject | None]
    objs_per_key = dict()
    for key in key_list:
        objs_per_key[key] = None
    async for row in await db_sess.stream(query):
        objs_per_key[key_getter(row)] = obj_type.from_row(graph_ctx, row)
    return [*objs_per_key.values()]


async def batch_result_in_scalar_stream(
    graph_ctx: GraphQueryContext,
    db_sess: SASession,
    query: sa.sql.Select[Any],
    obj_type: type[T_SQLBasedGQLObject],
    key_list: Iterable[T_Key],
    key_getter: Callable[[Row[Any]], T_Key],
) -> Sequence[T_SQLBasedGQLObject | None]:
    """
    A batched query adaptor for (key -> item) resolving patterns.
    stream the result scalar in async session.
    """
    objs_per_key: dict[T_Key, T_SQLBasedGQLObject | None]
    objs_per_key = {}
    for key in key_list:
        objs_per_key[key] = None
    async for row in await db_sess.stream_scalars(query):
        objs_per_key[key_getter(row)] = obj_type.from_row(graph_ctx, row)
    return [*objs_per_key.values()]


async def batch_multiresult_in_session(
    graph_ctx: GraphQueryContext,
    db_sess: SASession,
    query: sa.sql.Select[Any],
    obj_type: type[T_SQLBasedGQLObject],
    key_list: Iterable[T_Key],
    key_getter: Callable[[Row[Any]], T_Key],
) -> Sequence[Sequence[T_SQLBasedGQLObject]]:
    """
    A batched query adaptor for (key -> [item]) resolving patterns.
    stream the result in async session.
    """
    objs_per_key: dict[T_Key, list[T_SQLBasedGQLObject]]
    objs_per_key = dict()
    for key in key_list:
        objs_per_key[key] = list()
    async for row in await db_sess.stream(query):
        objs_per_key[key_getter(row)].append(
            obj_type.from_row(graph_ctx, row),
        )
    return [*objs_per_key.values()]


async def batch_multiresult_in_scalar_stream(
    graph_ctx: GraphQueryContext,
    db_sess: SASession,
    query: sa.sql.Select[Any],
    obj_type: type[T_SQLBasedGQLObject],
    key_list: Iterable[T_Key],
    key_getter: Callable[[Row[Any]], T_Key],
) -> Sequence[Sequence[T_SQLBasedGQLObject]]:
    """
    A batched query adaptor for (key -> [item]) resolving patterns.
    stream the result in async session.
    """
    objs_per_key: dict[T_Key, list[T_SQLBasedGQLObject]]
    objs_per_key = dict()
    for key in key_list:
        objs_per_key[key] = list()
    async for row in await db_sess.stream_scalars(query):
        objs_per_key[key_getter(row)].append(
            obj_type.from_row(graph_ctx, row),
        )
    return [*objs_per_key.values()]


def privileged_query(required_role: UserRole) -> Callable[..., Any]:
    def wrap(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        async def wrapped(
            root: Any,
            info: graphene.ResolveInfo,
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            from ai.backend.manager.models.user import UserRole

            ctx: GraphQueryContext = info.context
            if ctx.user["role"] != UserRole.SUPERADMIN:
                raise GenericForbidden("superadmin privilege required")
            return await func(root, info, *args, **kwargs)

        return wrapped

    return wrap


def scoped_query(
    *,
    autofill_user: bool = False,
    user_key: str = "access_key",
) -> Callable[..., Any]:
    """
    Prepends checks for domain/group/user access rights depending
    on the client's user and keypair information.

    :param autofill_user: When the *user_key* is not specified,
        automatically fills out the user data with the current
        user who is makeing the API request.
    :param user_key: The key used for storing user identification value
        in the keyword arguments.
    """

    def wrap(resolve_func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(resolve_func)
        async def wrapped(
            root: Any,
            info: graphene.ResolveInfo,
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            from ai.backend.manager.models.user import UserRole

            ctx: GraphQueryContext = info.context
            client_role = ctx.user["role"]
            if user_key == "access_key":
                client_user_id = ctx.access_key
            elif user_key == "email":
                client_user_id = ctx.user["email"]
            else:
                client_user_id = ctx.user["uuid"]
            client_domain = ctx.user["domain_name"]
            domain_name = kwargs.get("domain_name")
            group_id = kwargs.get("group_id") or kwargs.get("project_id")
            user_id = kwargs.get(user_key)
            if client_role == UserRole.SUPERADMIN:
                if autofill_user:
                    if user_id is None:
                        user_id = client_user_id
            elif client_role == UserRole.ADMIN:
                if domain_name is not None and domain_name != client_domain:
                    raise GenericForbidden
                domain_name = client_domain
                if group_id is not None:
                    # TODO: check if the group is a member of the domain
                    pass
                if autofill_user:
                    if user_id is None:
                        user_id = client_user_id
            elif client_role == UserRole.USER:
                if domain_name is not None and domain_name != client_domain:
                    raise GenericForbidden
                domain_name = client_domain
                if group_id is not None:
                    # TODO: check if the group is a member of the domain
                    # TODO: check if the client is a member of the group
                    pass
                if user_id is not None and user_id != client_user_id:
                    raise GenericForbidden
                user_id = client_user_id
            else:
                raise InvalidAPIParameters("Unknown client role")
            kwargs["domain_name"] = domain_name
            if group_id is not None:
                kwargs["group_id"] = group_id
            kwargs[user_key] = user_id
            return await resolve_func(root, info, *args, **kwargs)

        return wrapped

    return wrap


def privileged_mutation(
    required_role: UserRole, target_func: Callable[..., Any] | None = None
) -> Callable[..., Any]:
    def wrap(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        async def wrapped(
            cls: type, root: Any, info: graphene.ResolveInfo, *args: Any, **kwargs: Any
        ) -> Any:
            from ai.backend.manager.models.group import groups  # , association_groups_users
            from ai.backend.manager.models.user import UserRole

            ctx: GraphQueryContext = info.context
            permitted = False
            if required_role == UserRole.SUPERADMIN:
                if ctx.user["role"] == required_role:
                    permitted = True
            elif required_role == UserRole.ADMIN:
                if ctx.user["role"] == UserRole.SUPERADMIN:
                    permitted = True
                elif ctx.user["role"] == UserRole.USER:
                    permitted = False
                else:
                    if target_func is None:
                        return cls(False, "misconfigured privileged mutation: no target_func", None)
                    target_domain, target_group = target_func(*args, **kwargs)
                    if target_domain is None and target_group is None:
                        return cls(
                            False,
                            "misconfigured privileged mutation: "
                            "both target_domain and target_group missing",
                            None,
                        )
                    permit_chains = []
                    if target_domain is not None:
                        if ctx.user["domain_name"] == target_domain:
                            permit_chains.append(True)
                    if target_group is not None:
                        async with ctx.db.begin() as conn:
                            # check if the group is part of the requester's domain.
                            query = groups.select().where(
                                (groups.c.id == target_group)
                                & (groups.c.domain_name == ctx.user["domain_name"]),
                            )
                            result = await conn.execute(query)
                            if result.rowcount > 0:
                                permit_chains.append(True)
                            # TODO: check the group permission if implemented
                            # query = (
                            #     association_groups_users.select()
                            #     .where(association_groups_users.c.group_id == target_group)
                            # )
                            # result = await conn.execute(query)
                            # if result.rowcount > 0:
                            #     permit_chains.append(True)
                    permitted = all(permit_chains) if permit_chains else False
            elif required_role == UserRole.USER:
                permitted = True
            # assuming that mutation result objects has 2 or 3 fields:
            # success(bool), message(str) - usually for delete mutations
            # success(bool), message(str), item(object)
            if permitted:
                return await func(cls, root, info, *args, **kwargs)
            return cls(False, f"no permission to execute {info.path[0]}")

        return wrapped

    return wrap


ResultType = TypeVar("ResultType", bound=graphene.ObjectType)
ItemType = TypeVar("ItemType", bound=graphene.ObjectType)


async def gql_mutation_wrapper(
    result_cls: type[ResultType], _do_mutate: Callable[[], Awaitable[ResultType]]
) -> ResultType:
    try:
        return await execute_with_retry(_do_mutate)
    except sa.exc.IntegrityError as e:
        log.warning("gql_mutation_wrapper(): integrity error ({})", repr(e))
        return cast(ResultType, result_cls(False, f"integrity error: {e}"))
    except sa.exc.StatementError as e:
        log.warning(
            "gql_mutation_wrapper(): statement error ({})\n{}", repr(e), e.statement or "(unknown)"
        )
        orig_exc = e.orig
        return cast(ResultType, result_cls(False, str(orig_exc), None))
    except (TimeoutError, asyncio.CancelledError):
        raise
    except Exception as e:
        log.exception("gql_mutation_wrapper(): other error")
        return cast(ResultType, result_cls(False, f"unexpected error: {e}"))


async def simple_db_mutate[ResultType: graphene.ObjectType](
    result_cls: type[ResultType],
    graph_ctx: GraphQueryContext,
    mutation_query: sa.sql.Update
    | sa.sql.Insert
    | sa.sql.Delete
    | Callable[[], sa.sql.Update | sa.sql.Insert | sa.sql.Delete],
    *,
    pre_func: Callable[[SAConnection], Awaitable[None]] | None = None,
    post_func: Callable[[SAConnection, Result[Any]], Awaitable[None]] | None = None,
) -> ResultType:
    """
    Performs a database mutation based on the given
    :class:`sqlalchemy.sql.Update`, :class:`sqlalchemy.sql.Insert`, or
    :class:`sqlalchemy.sql.Delete` query, and return the wrapped result as the
    GraphQL object type given as **result_cls**.
    **result_cls** should have two initialization arguments: success (bool)
    and message (str).

    See details about the arguments in :func:`simple_db_mutate_returning_item`.
    """

    async def _do_mutate() -> ResultType:
        async with graph_ctx.db.begin() as conn:
            if pre_func:
                await pre_func(conn)
            _query = mutation_query() if callable(mutation_query) else mutation_query
            result = await conn.execute(_query)
            if post_func:
                await post_func(conn, result)
        if result.rowcount > 0:
            return cast(ResultType, result_cls(True, "success"))
        return cast(ResultType, result_cls(False, f"no matching {result_cls.__name__.lower()}"))

    return await gql_mutation_wrapper(result_cls, _do_mutate)


async def simple_db_mutate_returning_item[
    ResultType: graphene.ObjectType,
    ItemType: graphene.ObjectType,
](
    result_cls: type[ResultType],
    graph_ctx: GraphQueryContext,
    mutation_query: sa.sql.Update | sa.sql.Insert | Callable[[], sa.sql.Update | sa.sql.Insert],
    *,
    item_cls: type[ItemType],
    pre_func: Callable[[SAConnection], Awaitable[None]] | None = None,
    post_func: Callable[[SAConnection, Result[Any]], Awaitable[Row[Any]]] | None = None,
) -> ResultType:
    """
    Performs a database mutation based on the given
    :class:`sqlalchemy.sql.Update` or :class:`sqlalchemy.sql.Insert` query,
    and return the wrapped result as the GraphQL object type given as **result_cls**
    and the inserted/updated row wrapped as its 3rd argument in **item_cls**.

    If mutation_query uses external variable updated by pre_func, you should wrap the query
    with lambda so that its parameters are re-evaluated when the transaction is retried.

    :param result_cls: The GraphQL Object Type used to wrap the result.
        It should have two initialization arguments: success (bool),
        message (str), and the item (ItemType).
    :param graph_ctx: The common context that provides the reference to the database engine
        and other stuffs required to resolve the GraphQL query.
    :param mutation_query: A SQLAlchemy query object.
    :param item_cls: The GraphQL Object Type used to wrap the returned row from the mutation query.
    :param pre_func: An extra function that is executed before the mutation query, where the caller
        may perform additional database queries.
    :param post_func: An extra function that is executed after the mutation query, where the caller
        may perform additional database queries.  Note that it **MUST return the returned row
        from the given mutation result**, because the result object could be fetched only one
        time due to its cursor-like nature.
    """

    async def _do_mutate() -> ResultType:
        async with graph_ctx.db.begin() as conn:
            if pre_func:
                await pre_func(conn)
            _query = mutation_query() if callable(mutation_query) else mutation_query
            _query = _query.returning(_query.table)
            result = await conn.execute(_query)
            row: Row[Any] | None
            if post_func:
                row = await post_func(conn, result)
            else:
                row = result.first()
            if result.rowcount > 0 and row is not None:
                return cast(
                    ResultType, result_cls(True, "success", item_cls.from_row(graph_ctx, row))
                )
            return cast(
                ResultType, result_cls(False, f"no matching {result_cls.__name__.lower()}", None)
            )

    return await gql_mutation_wrapper(result_cls, _do_mutate)


def set_if_set(
    src: Any,
    target: MutableMapping[str, Any],
    name: str,
    *,
    clean_func: Callable[[Any], Any] | None = None,
    target_key: str | None = None,
) -> None:
    """
    Set the target dict with only non-undefined keys and their values
    from a Graphene's input object.
    (server-side function)
    """
    match src:
        case Mapping():
            v = src.get(name, Undefined)
        case _:
            v = getattr(src, name)
    # NOTE: unset optional fields are passed as graphql.Undefined.
    if v is not Undefined:
        if callable(clean_func):
            target[target_key or name] = clean_func(v)
        else:
            target[target_key or name] = v


def orm_set_if_set(
    src: object,
    target: MutableMapping[str, Any],
    name: str,
    *,
    clean_func: Callable[[Any], Any] | None = None,
    target_key: str | None = None,
) -> None:
    """
    Set the target ORM row object with only non-undefined keys and their values
    from a Graphene's input object.
    (server-side function)
    """
    v = getattr(src, name)
    # NOTE: unset optional fields are passed as graphql.Undefined.
    if v is not Undefined:
        if callable(clean_func):
            setattr(target, target_key or name, clean_func(v))
        else:
            setattr(target, target_key or name, v)


def filter_gql_undefined[T](val: T, *, default_value: T | None = None) -> T | None:
    if val is Undefined:
        return default_value
    return val


class InferenceSessionError(graphene.ObjectType):  # type: ignore[misc]
    class InferenceSessionErrorInfo(graphene.ObjectType):  # type: ignore[misc]
        src = graphene.String(required=True)
        name = graphene.String(required=True)
        repr = graphene.String(required=True)

    session_id = graphene.UUID()

    errors = graphene.List(graphene.NonNull(InferenceSessionErrorInfo), required=True)


class AsyncPaginatedConnectionField(AsyncListConnectionField):
    def __init__(self, type: type | str, *args: Any, **kwargs: Any) -> None:
        kwargs.setdefault("filter", graphene.String())
        kwargs.setdefault("order", graphene.String())
        kwargs.setdefault("offset", graphene.Int())
        super().__init__(type, *args, **kwargs)


PaginatedConnectionField = AsyncPaginatedConnectionField


class ConnectionArgs(NamedTuple):
    cursor: str | None
    pagination_order: ConnectionPaginationOrder | None
    requested_page_size: int


def validate_connection_args(
    *,
    after: str | None = None,
    first: int | None = None,
    before: str | None = None,
    last: int | None = None,
) -> ConnectionArgs:
    """
    Validate arguments used for GraphQL relay connection, and determine pagination ordering, cursor and page size.
    It is not allowed to use arguments for forward pagination and arguments for backward pagination at the same time.
    """
    order: ConnectionPaginationOrder | None = None
    cursor: str | None = None
    requested_page_size: int | None = None

    if after is not None:
        order = ConnectionPaginationOrder.FORWARD
        cursor = after
    if first is not None:
        if first < 0:
            raise ValueError("Argument 'first' must be a non-negative integer.")
        order = ConnectionPaginationOrder.FORWARD
        requested_page_size = first

    if before is not None:
        if order is ConnectionPaginationOrder.FORWARD:
            raise ValueError(
                "Can only paginate with single direction, forwards or backwards. Please set only"
                " one of (after, first) or (before, last)."
            )
        order = ConnectionPaginationOrder.BACKWARD
        cursor = before
    if last is not None:
        if last < 0:
            raise ValueError("Argument 'last' must be a non-negative integer.")
        if order is ConnectionPaginationOrder.FORWARD:
            raise ValueError(
                "Can only paginate with single direction, forwards or backwards. Please set only"
                " one of (after, first) or (before, last)."
            )
        order = ConnectionPaginationOrder.BACKWARD
        requested_page_size = last

    if requested_page_size is None:
        requested_page_size = DEFAULT_PAGE_SIZE

    return ConnectionArgs(cursor, order, requested_page_size)


@dataclass
class _StmtWithConditions:
    stmt: sa.sql.Select[Any]
    conditions: list[WhereClauseType]


def _apply_ordering(
    stmt: sa.sql.Select[Any],
    id_column: sa.Column[Any] | InstrumentedAttribute[Any],
    ordering_item_list: list[OrderingItem],
    pagination_order: ConnectionPaginationOrder | None,
) -> sa.sql.Select[Any]:
    """
    Apply ORDER BY clauses for cursor-based pagination with deterministic ordering.
    This function applies the user-specified ordering columns first, then adds the id column
    as the last tiebreaker to ensure deterministic ordering (required for stable cursor pagination).
    """
    id_ordering_item: OrderingItem
    set_ordering: Callable[
        [
            sa.Column[Any] | InstrumentedAttribute[Any] | sa.sql.elements.KeyedColumnElement[Any],
            OrderDirection,
        ],
        Any,
    ]
    match pagination_order:
        case ConnectionPaginationOrder.FORWARD | None:
            # Default ordering by id column (ascending for forward pagination)
            id_ordering_item = OrderingItem(id_column, OrderDirection.ASC)

            def set_ordering(
                col: sa.Column[Any]
                | InstrumentedAttribute[Any]
                | sa.sql.elements.KeyedColumnElement[Any],
                direction: OrderDirection,
            ) -> Any:
                return col.asc() if direction == OrderDirection.ASC else col.desc()

        case ConnectionPaginationOrder.BACKWARD:
            # Default ordering by id column (descending for backward pagination)
            id_ordering_item = OrderingItem(id_column, OrderDirection.DESC)

            # Reverse ordering direction for backward pagination
            def set_ordering(
                col: sa.Column[Any]
                | InstrumentedAttribute[Any]
                | sa.sql.elements.KeyedColumnElement[Any],
                direction: OrderDirection,
            ) -> Any:
                return col.desc() if direction == OrderDirection.ASC else col.asc()

    # Apply ordering to stmt (id column should be applied last for deterministic ordering)
    for col, direction in [*ordering_item_list, id_ordering_item]:
        stmt = stmt.order_by(set_ordering(col, direction))

    return stmt


def _apply_filter_conditions(
    stmt: sa.sql.Select[Any],
    orm_class: type[Any],
    filter_expr: FilterExprArg,
) -> _StmtWithConditions:
    """
    Filter conditions should be applied to both the main query statement and the count statement,
    as they define which items are included in the result set regardless of pagination.
    """
    filter_conditions: list[WhereClauseType] = []
    condition_parser = filter_expr.parser
    filter_cond = condition_parser.parse_filter(orm_class, filter_expr.expr)
    filter_conditions.append(filter_cond)
    stmt = stmt.where(filter_cond)

    return _StmtWithConditions(stmt, filter_conditions)


def _apply_cursor_pagination(
    info: graphene.ResolveInfo,
    stmt: sa.sql.Select[Any],
    id_column: sa.Column[Any] | InstrumentedAttribute[Any],
    ordering_item_list: list[OrderingItem],
    cursor_id: str,
    pagination_order: ConnectionPaginationOrder | None,
) -> _StmtWithConditions:
    """
    Apply cursor-based pagination WHERE conditions to the statement.
    """
    cursor_conditions: list[WhereClauseType] = []
    _, cursor_row_id_str = AsyncNode.resolve_global_id(info, cursor_id)

    cursor_row_id: UUID | str
    try:
        cursor_row_id = uuid.UUID(cursor_row_id_str)
    except (ValueError, AttributeError):
        # Fall back to string if not a valid UUID (for other ID types)
        cursor_row_id = cursor_row_id_str

    def subq_to_condition(
        column_to_be_compared: sa.Column[Any]
        | InstrumentedAttribute[Any]
        | sa.sql.elements.KeyedColumnElement[Any],
        subquery: ScalarSelect[Any],
        direction: OrderDirection,
    ) -> WhereClauseType:
        """Generate cursor condition for a specific ordering column.

        This handles cursor conditions when explicit order_expr is provided.
        For example, if ordering by "created_at DESC", this ensures we only get items
        where created_at < cursor_created_at, or where created_at = cursor_created_at but id < cursor_id.
        """
        cond: WhereClauseType
        condition_when_same_with_subq: WhereClauseType
        match pagination_order:
            case ConnectionPaginationOrder.FORWARD | None:
                if direction == OrderDirection.ASC:
                    cond = column_to_be_compared > subquery
                else:
                    cond = column_to_be_compared < subquery

                # Comparing ID field - The direction of inequality sign is not affected by `direction` argument here
                # because the ordering direction of ID field is always determined by `pagination_order` only.
                condition_when_same_with_subq = (column_to_be_compared == subquery) & (
                    id_column > cursor_row_id
                )
            case ConnectionPaginationOrder.BACKWARD:
                if direction == OrderDirection.ASC:
                    cond = column_to_be_compared < subquery
                else:
                    cond = column_to_be_compared > subquery
                condition_when_same_with_subq = (column_to_be_compared == subquery) & (
                    id_column < cursor_row_id
                )

        return cond | condition_when_same_with_subq

    # Add cursor conditions for explicit ordering columns (if any)
    for col, direction in ordering_item_list:
        subq = sa.select(col).where(id_column == cursor_row_id).scalar_subquery()
        cursor_conditions.append(subq_to_condition(col, subq, direction))

    # Add id-based cursor WHERE condition ONLY when no explicit ordering is provided.
    # This is CRITICAL for pagination to work when no explicit order_expr is provided.
    # When ordering_item_list is not empty, the id condition is already embedded
    # in the ordering cursor conditions above (via condition_when_same_with_subq).
    if not ordering_item_list:
        match pagination_order:
            case ConnectionPaginationOrder.FORWARD | None:
                cursor_conditions.append(id_column > cursor_row_id)
            case ConnectionPaginationOrder.BACKWARD:
                cursor_conditions.append(id_column < cursor_row_id)

    for cond in cursor_conditions:
        stmt = stmt.where(cond)

    return _StmtWithConditions(stmt, cursor_conditions)


def _build_sql_stmt_from_connection_args(
    info: graphene.ResolveInfo,
    orm_class: type[Any],
    id_column: sa.Column[Any] | InstrumentedAttribute[Any],
    filter_expr: FilterExprArg | None = None,
    order_expr: OrderExprArg | None = None,
    *,
    connection_args: ConnectionArgs,
) -> tuple[sa.sql.Select[Any], sa.sql.Select[Any], list[WhereClauseType]]:
    stmt = sa.select(orm_class)
    count_stmt = sa.select(sa.func.count()).select_from(orm_class)

    cursor_id, pagination_order, requested_page_size = connection_args

    # Parse explicit ordering from order_expr parameter (if provided)
    ordering_item_list: list[OrderingItem] = []
    if order_expr is not None:
        ordering_item_list = order_expr.parser.parse_order(orm_class, order_expr.expr)

    # Apply ORDER BY for cursor-based pagination
    stmt = _apply_ordering(stmt, id_column, ordering_item_list, pagination_order)

    # Apply filter conditions
    filter_conditions = []
    if filter_expr is not None:
        filter_result = _apply_filter_conditions(stmt, orm_class, filter_expr)
        stmt = filter_result.stmt
        for cond in filter_result.conditions:
            count_stmt = count_stmt.where(cond)
        filter_conditions = filter_result.conditions

    # Apply cursor pagination WHERE conditions (to stmt only)
    cursor_conditions = []
    if cursor_id is not None:
        cursor_result = _apply_cursor_pagination(
            info, stmt, id_column, ordering_item_list, cursor_id, pagination_order
        )
        stmt = cursor_result.stmt
        cursor_conditions = cursor_result.conditions

    # Apply LIMIT (to stmt only)
    if requested_page_size is not None:
        stmt = stmt.limit(requested_page_size + 1)

    return stmt, count_stmt, [*filter_conditions, *cursor_conditions]


def _build_sql_stmt_from_sql_arg(
    info: graphene.ResolveInfo,
    orm_class: type[Any],
    id_column: sa.Column[Any] | InstrumentedAttribute[Any],
    filter_expr: FilterExprArg | None = None,
    order_expr: OrderExprArg | None = None,
    *,
    limit: int,
    offset: int | None = None,
) -> tuple[sa.sql.Select[Any], sa.sql.Select[Any], list[WhereClauseType]]:
    stmt = sa.select(orm_class)
    count_stmt = sa.select(sa.func.count()).select_from(orm_class)
    conditions: list[WhereClauseType] = []

    if order_expr is not None:
        parser = order_expr.parser
        stmt = parser.append_ordering(stmt, order_expr.expr)

    # default order_by id column
    stmt = stmt.order_by(id_column.asc())

    if filter_expr is not None:
        condition_parser = filter_expr.parser
        conditions.append(condition_parser.parse_filter(orm_class, filter_expr.expr))

    if limit is not None:
        stmt = stmt.limit(limit)

    if offset is not None:
        stmt = stmt.offset(offset)
    for cond in conditions:
        stmt = stmt.where(cond)
        count_stmt = count_stmt.where(cond)
    return stmt, count_stmt, conditions


class GraphQLConnectionSQLInfo(NamedTuple):
    sql_stmt: sa.sql.Select[Any]
    sql_count_stmt: sa.sql.Select[Any]
    sql_conditions: list[WhereClauseType]
    cursor: str | None
    pagination_order: ConnectionPaginationOrder | None
    requested_page_size: int


class FilterExprArg(NamedTuple):
    expr: str
    parser: QueryFilterParser


class OrderExprArg(NamedTuple):
    expr: str
    parser: QueryOrderParser


def generate_sql_info_for_gql_connection(
    info: graphene.ResolveInfo,
    orm_class: type[Any],
    id_column: sa.Column[Any] | InstrumentedAttribute[Any],
    filter_expr: FilterExprArg | None = None,
    order_expr: OrderExprArg | None = None,
    offset: int | None = None,
    after: str | None = None,
    first: int | None = None,
    before: str | None = None,
    last: int | None = None,
) -> GraphQLConnectionSQLInfo:
    """
    Get GraphQL arguments and generate SQL query statement, cursor that points an id of a node, pagination order, and page size.
    If `offset` is None, return SQL query parsed from GraphQL Connection spec arguments.
    Else, return normally paginated SQL query and `first` is used as SQL limit.
    """

    if offset is None:
        connection_args = validate_connection_args(
            after=after, first=first, before=before, last=last
        )
        stmt, count_stmt, conditions = _build_sql_stmt_from_connection_args(
            info,
            orm_class,
            id_column,
            filter_expr,
            order_expr,
            connection_args=connection_args,
        )
        ret = GraphQLConnectionSQLInfo(
            stmt,
            count_stmt,
            conditions,
            connection_args.cursor,
            connection_args.pagination_order,
            connection_args.requested_page_size,
        )
    else:
        page_size = first if first is not None else DEFAULT_PAGE_SIZE
        stmt, count_stmt, conditions = _build_sql_stmt_from_sql_arg(
            info,
            orm_class,
            id_column,
            filter_expr,
            order_expr,
            limit=page_size,
            offset=offset,
        )
        ret = GraphQLConnectionSQLInfo(stmt, count_stmt, conditions, None, None, page_size)

    ctx: GraphQueryContext = info.context
    max_page_size = ctx.config_provider.config.api.max_gql_connection_page_size
    if max_page_size is not None and ret.requested_page_size > max_page_size:
        raise ValueError(
            f"Cannot fetch a page larger than {max_page_size}. "
            "Set 'first' or 'last' to a smaller integer."
        )
    return ret



"""Domain Usage Bucket query resolvers."""

from __future__ import annotations

import strawberry
from aiohttp import web
from strawberry import Info

from ai.backend.common.contexts.user import current_user
from ai.backend.manager.api.gql.resource_usage.fetcher import fetch_domain_usage_buckets
from ai.backend.manager.api.gql.resource_usage.types import (
    DomainUsageBucketConnection,
    DomainUsageBucketFilter,
    DomainUsageBucketOrderBy,
)
from ai.backend.manager.api.gql.types import ResourceGroupDomainScope, StrawberryGQLContext
from ai.backend.manager.api.gql.utils import check_admin_only

# Admin APIs


@strawberry.field(description="Added in 26.2.0. List domain usage buckets (admin only).")  # type: ignore[misc]
async def admin_domain_usage_buckets(
    info: Info[StrawberryGQLContext],
    filter: DomainUsageBucketFilter | None = None,
    order_by: list[DomainUsageBucketOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> DomainUsageBucketConnection | None:
    """Search domain usage buckets with pagination (admin only)."""
    check_admin_only()

    return await fetch_domain_usage_buckets(
        info=info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


# Resource Group Scoped APIs


@strawberry.field(  # type: ignore[misc]
    description=(
        "Added in 26.2.0. List domain usage buckets within resource group scope. "
        "This API is not yet implemented."
    )
)
async def rg_domain_usage_buckets(
    info: Info[StrawberryGQLContext],
    scope: ResourceGroupDomainScope,
    filter: DomainUsageBucketFilter | None = None,
    order_by: list[DomainUsageBucketOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> DomainUsageBucketConnection | None:
    """Search domain usage buckets within resource group scope."""
    raise NotImplementedError("rg_domain_usage_buckets is not yet implemented")


# Legacy APIs (deprecated)


@strawberry.field(  # type: ignore[misc]
    description="Added in 26.1.0. List domain usage buckets (superadmin only).",
    deprecation_reason=(
        "Use admin_domain_usage_buckets instead. "
        "This API will be removed after v26.3.0. See BEP-1041 for migration guide."
    ),
)
async def domain_usage_buckets(
    info: Info[StrawberryGQLContext],
    filter: DomainUsageBucketFilter | None = None,
    order_by: list[DomainUsageBucketOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> DomainUsageBucketConnection | None:
    """Search domain usage buckets with pagination."""
    me = current_user()
    if me is None or not me.is_superadmin:
        raise web.HTTPForbidden(reason="Only superadmin can access usage bucket data.")

    return await fetch_domain_usage_buckets(
        info=info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )



from __future__ import annotations

import uuid

from ai.backend.common.contexts.user import current_user
from ai.backend.manager.data.model_serving.types import (
    EndpointAccessValidationData,
)
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.services.model_serving.exceptions import (
    GenericForbidden,
    InvalidAPIParameters,
)
from ai.backend.manager.utils import check_if_requester_is_eligible_to_act_as_target_user_uuid


async def verify_user_access_scopes(db: ExtendedAsyncSAEngine, owner_uuid: uuid.UUID) -> None:
    user_data = current_user()
    if user_data is None or user_data.is_authorized is False:
        raise GenericForbidden("Only authorized requests may have access key scopes.")
    if owner_uuid is None or owner_uuid == user_data.user_id:
        return
    async with db.begin_readonly() as conn:
        try:
            await check_if_requester_is_eligible_to_act_as_target_user_uuid(
                conn,
                user_data.role,
                user_data.domain_name,
                owner_uuid,
            )
            return
        except ValueError as e:
            raise InvalidAPIParameters(str(e)) from e
        except RuntimeError as e:
            raise GenericForbidden(str(e)) from e


def validate_endpoint_access(
    validation_data: EndpointAccessValidationData,
) -> bool:
    """Validate user access to endpoint based on role.

    Returns True if the user has access to the endpoint, False otherwise.

    Access rules:
    - SUPERADMIN: Full access to all endpoints
    - ADMIN: Access to endpoints in their domain, except those owned by SUPERADMIN
    - USER/others: Access only to endpoints they own
    """
    user_data = current_user()
    if user_data is None:
        return False

    if validation_data.session_owner_id is None:
        return True

    match user_data.role:
        case UserRole.SUPERADMIN:
            return True
        case UserRole.ADMIN:
            # ADMIN cannot access SUPERADMIN's resources
            if validation_data.session_owner_role == UserRole.SUPERADMIN:
                return False
            return validation_data.domain == user_data.domain_name
        case _:
            return validation_data.session_owner_id == user_data.user_id



"""User GraphQL API package.

Added in 26.2.0. Provides structured user management API with typed fields
replacing JSON scalars and organized into logical field groups.
"""

from .resolver import (
    # Mutations
    admin_bulk_create_users,
    admin_create_user,
    admin_delete_user,
    admin_delete_users,
    admin_purge_user,
    admin_purge_users,
    admin_update_user,
    # Queries
    admin_user_v2,
    admin_users_v2,
    domain_users_v2,
    my_user_v2,
    project_users_v2,
    update_user,
)

__all__ = [
    # Queries
    "admin_user_v2",
    "admin_users_v2",
    "domain_users_v2",
    "my_user_v2",
    "project_users_v2",
    # Mutations
    "admin_create_user",
    "admin_bulk_create_users",
    "admin_update_user",
    "update_user",
    "admin_delete_user",
    "admin_delete_users",
    "admin_purge_user",
    "admin_purge_users",
]



from __future__ import annotations

from dataclasses import dataclass, field
from enum import StrEnum
from typing import TYPE_CHECKING
from uuid import UUID

from ai.backend.manager.data.deployment.types import DeploymentInfo, RouteStatus

if TYPE_CHECKING:
    from ai.backend.manager.data.deployment.types import DeploymentInfoWithRoutes, RouteInfo


class DeploymentLifecycleType(StrEnum):
    CHECK_PENDING = "check_pending"
    CHECK_REPLICA = "check_replica"
    SCALING = "scaling"
    RECONCILE = "reconcile"
    DESTROYING = "destroying"


@dataclass
class DeploymentExecutionError:
    deployment_info: DeploymentInfo
    reason: str
    error_detail: str
    error_code: str | None = None


@dataclass
class DeploymentExecutionResult:
    """Result of a deployment execution operation."""

    successes: list[DeploymentInfo] = field(default_factory=list)
    errors: list[DeploymentExecutionError] = field(default_factory=list)
    skipped: list[DeploymentInfo] = field(default_factory=list)


@dataclass
class AutoScalingDecision:
    """Decision made by autoscaling evaluation."""

    should_scale: bool
    new_replica_count: int | None = None
    triggered_rule_id: UUID | None = None
    scaling_direction: str | None = None  # "up" or "down"
    reason: str | None = None


@dataclass
class RouteCreationSpec:
    """Specification for creating a new route and session."""

    endpoint_id: UUID
    endpoint_name: str
    traffic_ratio: float
    image_id: UUID
    resource_group: str
    domain: str
    project: UUID
    created_user: UUID
    session_owner: UUID
    model_mount_destination: str

    # Extension methods for DeploymentInfo compatibility
    @staticmethod
    def get_target_replicas_from_deployment(deployment_info: DeploymentInfo) -> int:
        """Get the target number of replicas for a DeploymentInfo."""
        # DeploymentInfo has replica_spec.replica_count
        return deployment_info.replica_spec.replica_count

    # Extension methods for DeploymentInfoWithRoutes compatibility
    @staticmethod
    def get_healthy_route_count_from_deployment(
        deployment_with_routes: DeploymentInfoWithRoutes,
    ) -> int:
        """Get the count of healthy routes."""
        return sum(
            1
            for route in deployment_with_routes.routes
            if route.status in {RouteStatus.HEALTHY, RouteStatus.PROVISIONING}
        )

    @staticmethod
    def get_routes_to_remove_from_deployment(
        deployment_with_routes: DeploymentInfoWithRoutes, target_count: int
    ) -> list[RouteInfo]:
        """Get routes that should be removed to reach target count."""
        healthy_routes = [
            r
            for r in deployment_with_routes.routes
            if r.status in {RouteStatus.HEALTHY, RouteStatus.PROVISIONING}
        ]
        current_count = len(healthy_routes)
        if current_count <= target_count:
            return []
        # Remove routes with lowest traffic ratio first
        sorted_routes = sorted(healthy_routes, key=lambda r: r.traffic_ratio)
        return sorted_routes[: current_count - target_count]



from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.scaling_group.types import ScalingGroupData

from .base import ScalingGroupAction


@dataclass(frozen=True)
class ResourceWeightInput:
    """Input for a single resource weight entry."""

    resource_type: str
    weight: Decimal | None  # None means delete


@dataclass
class UpdateFairShareSpecAction(ScalingGroupAction):
    """Action to update fair share spec for a scaling group.

    Supports partial updates - only provided fields are modified.
    Validates resource_weights against capacity and filters out
    resource types no longer available.
    """

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.RESOURCE_GROUP_FAIR_SHARE

    resource_group: str
    half_life_days: int | None = None
    lookback_days: int | None = None
    decay_unit_days: int | None = None
    default_weight: Decimal | None = None
    resource_weights: list[ResourceWeightInput] | None = None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE

    @override
    def entity_id(self) -> str:
        return self.resource_group


@dataclass
class UpdateFairShareSpecActionResult(BaseActionResult):
    """Result of updating fair share spec."""

    scaling_group: ScalingGroupData

    @override
    def entity_id(self) -> str | None:
        return self.scaling_group.name



"""
API-related exceptions.
"""

from __future__ import annotations

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class NotImplementedAPI(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/not-implemented"
    error_title = "This API is not implemented."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.NOT_IMPLEMENTED,
        )


class DeprecatedAPI(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/deprecated"
    error_title = "This API is deprecated."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.DEPRECATED,
        )


class InvalidAPIParameters(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-api-params"
    error_title = "Missing or invalid API parameters."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class GraphQLError(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/graphql-error"
    error_title = "GraphQL-generated error."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class RateLimitExceeded(BackendAIError, web.HTTPTooManyRequests):
    error_type = "https://api.backend.ai/probs/rate-limit-exceeded"
    error_title = "You have reached your API query rate limit."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class InvalidGraphQLParameters(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-graphql-params"
    error_title = "Invalid GraphQL parameters."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class InvalidCursor(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/invalid-cursor"
    error_title = "Invalid cursor format."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.API,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )



"""GraphQL types for DeploymentPolicy."""

from __future__ import annotations

from datetime import datetime
from typing import Self

import strawberry
from strawberry import ID
from strawberry.relay import Node, NodeID

from ai.backend.common.data.model_deployment.types import DeploymentStrategy
from ai.backend.manager.errors.deployment import InvalidDeploymentStrategySpec
from ai.backend.manager.models.deployment_policy import (
    BlueGreenSpec,
    DeploymentPolicyData,
    RollingUpdateSpec,
)

# Enum defined here to avoid circular import with deployment.py
DeploymentStrategyTypeGQL: type[DeploymentStrategy] = strawberry.enum(
    DeploymentStrategy,
    name="DeploymentStrategyType",
    description="Added in 25.19.0. This enum represents the deployment strategy type of a model deployment, indicating the strategy used for deployment.",
)

# ========== Output Types (Response) ==========


@strawberry.interface(
    name="DeploymentStrategySpec",
    description="Added in 25.19.0. Base interface for deployment strategy specifications.",
)
class DeploymentStrategySpecGQL:
    strategy: DeploymentStrategyTypeGQL


@strawberry.type(
    name="RollingUpdateStrategySpec",
    description="Added in 25.19.0. Rolling update strategy specification.",
)
class RollingUpdateStrategySpecGQL(DeploymentStrategySpecGQL):
    max_surge: int
    max_unavailable: int


@strawberry.type(
    name="BlueGreenStrategySpec",
    description="Added in 25.19.0. Blue-green deployment strategy specification.",
)
class BlueGreenStrategySpecGQL(DeploymentStrategySpecGQL):
    auto_promote: bool
    promote_delay_seconds: int


@strawberry.type(
    name="DeploymentPolicy",
    description="Added in 25.19.0. Deployment policy configuration.",
)
class DeploymentPolicyGQL(Node):
    id: NodeID[str]
    strategy_spec: DeploymentStrategySpecGQL
    rollback_on_failure: bool
    created_at: datetime
    updated_at: datetime

    @classmethod
    def from_data(cls, data: DeploymentPolicyData) -> Self:
        match data.strategy:
            case DeploymentStrategy.ROLLING:
                if not isinstance(data.strategy_spec, RollingUpdateSpec):
                    raise InvalidDeploymentStrategySpec(
                        "Expected RollingUpdateSpec for ROLLING strategy"
                    )
                return cls(
                    id=ID(str(data.id)),
                    strategy_spec=RollingUpdateStrategySpecGQL(
                        strategy=DeploymentStrategyTypeGQL.ROLLING,
                        max_surge=data.strategy_spec.max_surge,
                        max_unavailable=data.strategy_spec.max_unavailable,
                    ),
                    rollback_on_failure=data.rollback_on_failure,
                    created_at=data.created_at,
                    updated_at=data.updated_at,
                )
            case DeploymentStrategy.BLUE_GREEN:
                if not isinstance(data.strategy_spec, BlueGreenSpec):
                    raise InvalidDeploymentStrategySpec(
                        "Expected BlueGreenSpec for BLUE_GREEN strategy"
                    )
                return cls(
                    id=ID(str(data.id)),
                    strategy_spec=BlueGreenStrategySpecGQL(
                        strategy=DeploymentStrategyTypeGQL.BLUE_GREEN,
                        auto_promote=data.strategy_spec.auto_promote,
                        promote_delay_seconds=data.strategy_spec.promote_delay_seconds,
                    ),
                    rollback_on_failure=data.rollback_on_failure,
                    created_at=data.created_at,
                    updated_at=data.updated_at,
                )


# ========== Input Types ==========


@strawberry.input(
    name="RollingUpdateConfigInput",
    description="Added in 25.19.0. Configuration for rolling update strategy.",
)
class RollingUpdateConfigInputGQL:
    max_surge: int = 1
    max_unavailable: int = 0

    def to_spec(self) -> RollingUpdateSpec:
        return RollingUpdateSpec(
            max_surge=self.max_surge,
            max_unavailable=self.max_unavailable,
        )


@strawberry.input(
    name="BlueGreenConfigInput",
    description="Added in 25.19.0. Configuration for blue-green deployment strategy.",
)
class BlueGreenConfigInputGQL:
    auto_promote: bool = False
    promote_delay_seconds: int = 0

    def to_spec(self) -> BlueGreenSpec:
        return BlueGreenSpec(
            auto_promote=self.auto_promote,
            promote_delay_seconds=self.promote_delay_seconds,
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import UUID

from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience import (
    MetricArgs,
    MetricPolicy,
    Resilience,
    RetryArgs,
    RetryPolicy,
)
from ai.backend.common.resilience.policies.retry import BackoffStrategy
from ai.backend.manager.data.scaling_group.types import (
    ResourceInfo,
    ScalingGroupData,
    ScalingGroupListResult,
)
from ai.backend.manager.models.scaling_group import (
    ScalingGroupForDomainRow,
    ScalingGroupForKeypairsRow,
    ScalingGroupForProjectRow,
    ScalingGroupRow,
)
from ai.backend.manager.repositories.base import BatchQuerier
from ai.backend.manager.repositories.base.creator import BulkCreator, Creator
from ai.backend.manager.repositories.base.purger import BatchPurger, Purger
from ai.backend.manager.repositories.base.updater import Updater

from .db_source import ScalingGroupDBSource

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

__all__ = ("ScalingGroupRepository",)


scaling_group_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.SCALING_GROUP_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
            )
        ),
    ]
)


class ScalingGroupRepository:
    """Repository for scaling group-related data access."""

    _db_source: ScalingGroupDBSource

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
    ) -> None:
        self._db_source = ScalingGroupDBSource(db)

    @scaling_group_repository_resilience.apply()
    async def create_scaling_group(
        self,
        creator: Creator[ScalingGroupRow],
    ) -> ScalingGroupData:
        """Creates a new scaling group.

        Raises ScalingGroupConflict if a scaling group with the same name already exists.
        """
        return await self._db_source.create_scaling_group(creator)

    @scaling_group_repository_resilience.apply()
    async def search_scaling_groups(
        self,
        querier: BatchQuerier,
    ) -> ScalingGroupListResult:
        """Searches scaling groups with total count."""
        return await self._db_source.search_scaling_groups(querier=querier)

    @scaling_group_repository_resilience.apply()
    async def get_scaling_group_by_name(
        self,
        name: str,
    ) -> ScalingGroupData:
        """Get a single scaling group by name (primary key).

        Args:
            name: The name of the scaling group (primary key).

        Returns:
            ScalingGroupData for the requested scaling group.

        Raises:
            ScalingGroupNotFound: If the scaling group does not exist.
        """
        return await self._db_source.get_scaling_group_by_name(name=name)

    @scaling_group_repository_resilience.apply()
    async def purge_scaling_group(
        self,
        purger: Purger[ScalingGroupRow],
    ) -> ScalingGroupData:
        """Purges a scaling group and all related sessions and routes using a purger.

        Raises ScalingGroupNotFound if scaling group doesn't exist.
        """
        return await self._db_source.purge_scaling_group(purger)

    @scaling_group_repository_resilience.apply()
    async def update_scaling_group(
        self,
        updater: Updater[ScalingGroupRow],
    ) -> ScalingGroupData:
        """Updates an existing scaling group.

        Raises ScalingGroupNotFound if the scaling group does not exist.
        """
        return await self._db_source.update_scaling_group(updater)

    async def associate_scaling_group_with_domains(
        self,
        bulk_creator: BulkCreator[ScalingGroupForDomainRow],
    ) -> None:
        """Associates a scaling group with multiple domains."""
        await self._db_source.associate_scaling_group_with_domains(bulk_creator)

    async def disassociate_scaling_group_with_domains(
        self,
        purger: BatchPurger[ScalingGroupForDomainRow],
    ) -> None:
        """Disassociates a scaling group from multiple domains."""
        await self._db_source.disassociate_scaling_group_with_domains(purger)

    async def check_scaling_group_domain_association_exists(
        self,
        scaling_group: str,
        domain: str,
    ) -> bool:
        """Checks if a scaling group is associated with a domain."""
        return await self._db_source.check_scaling_group_domain_association_exists(
            scaling_group=scaling_group,
            domain=domain,
        )

    async def associate_scaling_group_with_keypairs(
        self,
        bulk_creator: BulkCreator[ScalingGroupForKeypairsRow],
    ) -> None:
        """Associates a scaling group with multiple keypairs."""
        await self._db_source.associate_scaling_group_with_keypairs(bulk_creator)

    async def disassociate_scaling_group_with_keypairs(
        self,
        purger: BatchPurger[ScalingGroupForKeypairsRow],
    ) -> None:
        """Disassociates a scaling group from multiple keypairs."""
        await self._db_source.disassociate_scaling_group_with_keypairs(purger)

    async def check_scaling_group_keypair_association_exists(
        self,
        scaling_group_name: str,
        access_key: str,
    ) -> bool:
        """Checks if a scaling group is associated with a keypair."""
        return await self._db_source.check_scaling_group_keypair_association_exists(
            scaling_group_name, access_key
        )

    async def associate_scaling_group_with_user_groups(
        self,
        bulk_creator: BulkCreator[ScalingGroupForProjectRow],
    ) -> None:
        """Associates a scaling group with multiple user groups (projects)."""
        await self._db_source.associate_scaling_group_with_user_groups(bulk_creator)

    async def disassociate_scaling_group_with_user_groups(
        self,
        purger: BatchPurger[ScalingGroupForProjectRow],
    ) -> None:
        """Disassociates a single scaling group from a user group (project)."""
        await self._db_source.disassociate_scaling_group_with_user_groups(purger)

    async def check_scaling_group_user_group_association_exists(
        self,
        scaling_group: str,
        user_group: UUID,
    ) -> bool:
        """Checks if a scaling group is associated with a user group (project)."""
        return await self._db_source.check_scaling_group_user_group_association_exists(
            scaling_group=scaling_group,
            user_group=user_group,
        )

    async def get_resource_info(
        self,
        scaling_group: str,
    ) -> ResourceInfo:
        """Get aggregated resource information for a scaling group.

        Args:
            scaling_group: The name of the scaling group.

        Returns:
            ResourceInfo containing capacity, used, and free resource metrics.

        Raises:
            ScalingGroupNotFound: If the scaling group does not exist.
        """
        return await self._db_source.get_resource_info(scaling_group)



import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.common.data.storage.types import ArtifactStorageType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.association.types import AssociationArtifactsStoragesData
from ai.backend.manager.services.artifact_revision.actions.base import ArtifactRevisionAction


@dataclass
class AssociateWithStorageAction(ArtifactRevisionAction):
    artifact_revision_id: uuid.UUID
    storage_namespace_id: uuid.UUID
    storage_type: ArtifactStorageType

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.ARTIFACT_REVISION_STORAGE_LINK

    @override
    def entity_id(self) -> str | None:
        return str(self.artifact_revision_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class AssociateWithStorageActionResult(BaseActionResult):
    result: AssociationArtifactsStoragesData

    @override
    def entity_id(self) -> str | None:
        return str(self.result.id)



import uuid

from pydantic import BaseModel

from ai.backend.common.api_handlers import BaseResponseModel
from ai.backend.common.data.artifact.types import (
    ArtifactRegistryType,
    CombinedDownloadProgress,
)
from ai.backend.manager.data.artifact.types import (
    ArtifactData,
    ArtifactDataWithRevisions,
    ArtifactDataWithRevisionsResponse,
    ArtifactRevisionReadme,
    ArtifactRevisionResponseData,
)


class SearchArtifactsResponse(BaseResponseModel):
    artifacts: list[ArtifactDataWithRevisionsResponse]


class ScanArtifactsResponse(BaseResponseModel):
    artifacts: list[ArtifactDataWithRevisionsResponse]


class ScanArtifactsSyncResponse(BaseResponseModel):
    artifacts: list[ArtifactDataWithRevisionsResponse]
    readme_data: dict[uuid.UUID, ArtifactRevisionReadme]


class DelegateScanArtifactsResponse(BaseResponseModel):
    artifacts: list[ArtifactDataWithRevisionsResponse]
    source_registry_id: uuid.UUID
    source_registry_type: ArtifactRegistryType
    readme_data: dict[uuid.UUID, ArtifactRevisionReadme]


class ScanArtifactModelsResponse(BaseResponseModel):
    artifacts: list[ArtifactDataWithRevisionsResponse]


class RetreiveArtifactModelsResponse(BaseResponseModel):
    artifacts: list[ArtifactDataWithRevisionsResponse]


class RetreiveArtifactModelResponse(BaseResponseModel):
    artifact: ArtifactDataWithRevisions


class CleanupArtifactsResponse(BaseResponseModel):
    artifact_revisions: list[ArtifactRevisionResponseData]


class CancelImportArtifactResponse(BaseResponseModel):
    artifact_revision: ArtifactRevisionResponseData


class ApproveArtifactRevisionResponse(BaseResponseModel):
    artifact_revision: ArtifactRevisionResponseData


class RejectArtifactRevisionResponse(BaseResponseModel):
    artifact_revision: ArtifactRevisionResponseData


class ArtifactRevisionImportTask(BaseResponseModel):
    task_id: str | None
    artifact_revision: ArtifactRevisionResponseData


class ImportArtifactsResponse(BaseResponseModel):
    tasks: list[ArtifactRevisionImportTask]


class DelegateImportArtifactsResponse(BaseResponseModel):
    tasks: list[ArtifactRevisionImportTask]


class UpdateArtifactResponse(BaseResponseModel):
    artifact: ArtifactData


class GetArtifactRevisionReadmeResponse(BaseResponseModel):
    readme: str | None


class VFSStorage(BaseModel):
    name: str
    base_path: str
    host: str


class GetVFSStorageResponse(BaseResponseModel):
    storage: VFSStorage


class ListVFSStorageResponse(BaseResponseModel):
    storages: list[VFSStorage]


class GetDownloadProgressResponse(BaseResponseModel):
    download_progress: CombinedDownloadProgress



"""Export repositories container."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Self

from .db_source import ExportDBSource
from .registry import ExportReportRegistry
from .repository import ExportRepository

if TYPE_CHECKING:
    from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class ExportRepositories:
    """Container for export-related repositories."""

    repository: ExportRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        """Create ExportRepositories from repository args."""
        db_source = ExportDBSource(db=args.db)
        registry = ExportReportRegistry.create_default()

        return cls(
            repository=ExportRepository(
                db_source=db_source,
                registry=registry,
            ),
        )



"""Action for updating route traffic status."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.deployment.types import RouteInfo, RouteTrafficStatus

from .base import RouteBaseAction


@dataclass
class UpdateRouteTrafficStatusAction(RouteBaseAction):
    """Action to update traffic status of a route."""

    route_id: UUID
    traffic_status: RouteTrafficStatus

    @override
    def entity_id(self) -> str | None:
        return str(self.route_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class UpdateRouteTrafficStatusActionResult(BaseActionResult):
    """Result of updating route traffic status."""

    route: RouteInfo

    @override
    def entity_id(self) -> str | None:
        return str(self.route.route_id)



"""add_is_public_flag_to_scaling_group

Revision ID: fe59ec332c07
Revises: bedd92de93af
Create Date: 2023-04-24 22:35:05.793777

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "fe59ec332c07"
down_revision = "bedd92de93af"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "scaling_groups",
        sa.Column("is_public", sa.Boolean(), server_default=sa.text("true"), nullable=False),
    )
    op.create_index(
        op.f("ix_scaling_groups_is_public"), "scaling_groups", ["is_public"], unique=False
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_scaling_groups_is_public"), table_name="scaling_groups")
    op.drop_column("scaling_groups", "is_public")
    # ### end Alembic commands ###



from .creators import (
    DeploymentHistoryCreatorSpec,
    KernelSchedulingHistoryCreatorSpec,
    RouteHistoryCreatorSpec,
    SessionSchedulingHistoryCreatorSpec,
)
from .repositories import SchedulingHistoryRepositories
from .repository import SchedulingHistoryRepository
from .types import (
    DeploymentHistorySearchScope,
    RouteHistorySearchScope,
    SessionSchedulingHistorySearchScope,
)

__all__ = (
    "DeploymentHistoryCreatorSpec",
    "DeploymentHistorySearchScope",
    "KernelSchedulingHistoryCreatorSpec",
    "RouteHistoryCreatorSpec",
    "RouteHistorySearchScope",
    "SchedulingHistoryRepositories",
    "SchedulingHistoryRepository",
    "SessionSchedulingHistoryCreatorSpec",
    "SessionSchedulingHistorySearchScope",
)



"""Data containers for model serving operations.

Note: These are data containers, not CreatorSpec implementations.
For row creation, use EndpointCreatorSpec from repositories/model_serving/creators.py
"""

from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from typing import TYPE_CHECKING

from pydantic import AnyUrl

from ai.backend.common.types import (
    AutoScalingMetricComparator,
    AutoScalingMetricSource,
    ClusterMode,
    RuntimeVariant,
)
from ai.backend.manager.data.model_serving.types import (
    ModelServicePrepareCtx,
    ServiceConfig,
)

if TYPE_CHECKING:
    from ai.backend.manager.data.deployment.types import ModelRevisionSpec


@dataclass
class ModelServiceCreator:
    """Data container for model service creation parameters.

    This is not a CreatorSpec - it's used to collect and pass service creation
    parameters in the service layer before creating an EndpointCreatorSpec.
    """

    service_name: str
    replicas: int
    image: str | None
    runtime_variant: RuntimeVariant
    architecture: str | None
    group_name: str
    domain_name: str
    cluster_size: int
    cluster_mode: ClusterMode
    open_to_public: bool
    config: ServiceConfig
    sudo_session_enabled: bool
    model_service_prepare_ctx: ModelServicePrepareCtx
    tag: str | None = None
    startup_command: str | None = None
    bootstrap_script: str | None = None
    callback_url: AnyUrl | None = None

    def with_revision(self, revision: ModelRevisionSpec) -> ModelServiceCreator:
        """Return a new creator with revision results applied."""
        overrided_service_config = dataclasses.replace(
            self.config,
            resources=dict(revision.resource_spec.resource_slots),
            environ=revision.execution.environ,
        )
        return dataclasses.replace(
            self,
            image=revision.image_identifier.canonical,
            architecture=revision.image_identifier.architecture,
            config=overrided_service_config,
        )


@dataclass
class EndpointAutoScalingRuleCreator:
    """Data container for endpoint auto scaling rule creation parameters.

    This is not a CreatorSpec - the repository extracts individual fields
    and creates the row directly without using CreatorSpec pattern.
    """

    metric_source: AutoScalingMetricSource
    metric_name: str
    threshold: str
    comparator: AutoScalingMetricComparator
    step_size: int
    cooldown_seconds: int
    min_replicas: int | None = None
    max_replicas: int | None = None



"""Deployment-specific recorder context for deployment coordinator."""

from __future__ import annotations

from uuid import UUID

from ai.backend.manager.sokovan.recorder.context import RecorderContext

# Deployment-specific type alias
DeploymentRecorderContext = RecorderContext[UUID]

__all__ = [
    "DeploymentRecorderContext",
]



from .creators import ErrorLogCreatorSpec
from .options import ErrorLogConditions
from .repositories import ErrorLogRepositories
from .repository import ErrorLogRepository

__all__ = (
    "ErrorLogConditions",
    "ErrorLogCreatorSpec",
    "ErrorLogRepositories",
    "ErrorLogRepository",
)



from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class UserNotFound(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/user-not-found"
    error_title = "The user does not exist."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class UserConflict(BackendAIError, web.HTTPConflict):
    error_type = "https://api.backend.ai/probs/user-conflict"
    error_title = "The user already exists."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.CONFLICT,
        )


class UserModificationBadRequest(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/user-modification-bad-request"
    error_title = "Failed to modify user due to bad request."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class UserCreationBadRequest(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/user-creation-bad-request"
    error_title = "Failed to create user due to bad request."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class UserCreationFailure(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/user-creation-failure"
    error_title = "Failed to create user."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class UserModificationFailure(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/user-modification-failure"
    error_title = "Failed to modify user."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.UPDATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class UserPurgeFailure(BackendAIError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/user-purge-failure"
    error_title = "Failed to purge user."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.HARD_DELETE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class KeyPairNotFound(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/keypair-not-found"
    error_title = "The key pair does not exist."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KEYPAIR,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class KeyPairForbidden(BackendAIError, web.HTTPForbidden):
    error_type = "https://api.backend.ai/probs/keypair-forbidden"
    error_title = "The key pair is not allowed to be used."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KEYPAIR,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.FORBIDDEN,
        )



from __future__ import annotations

import logging
import uuid
from pathlib import Path
from typing import TYPE_CHECKING

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, foreign, mapped_column, relationship

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.vfs_storage.types import VFSStorageData
from ai.backend.manager.models.base import (
    GUID,
    Base,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.association_artifacts_storages import (
        AssociationArtifactsStorageRow,
    )

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = ("VFSStorageRow",)


def _get_vfs_storage_association_artifact_join_cond() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.association_artifacts_storages import (
        AssociationArtifactsStorageRow,
    )

    return VFSStorageRow.id == foreign(AssociationArtifactsStorageRow.storage_namespace_id)


class VFSStorageRow(Base):  # type: ignore[misc]
    """
    Represents a VFS storage configuration.
    This model is used to store the details of VFS storage backends
    such as base paths, subpaths, and chunk sizes.
    """

    __tablename__ = "vfs_storages"

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    name: Mapped[str] = mapped_column("name", sa.String, index=True, unique=True, nullable=False)
    host: Mapped[str] = mapped_column("host", sa.String, nullable=False)
    base_path: Mapped[str] = mapped_column("base_path", sa.String, nullable=False)

    association_artifacts_storages_rows: Mapped[list[AssociationArtifactsStorageRow]] = (
        relationship(
            "AssociationArtifactsStorageRow",
            back_populates="vfs_storage_row",
            primaryjoin=_get_vfs_storage_association_artifact_join_cond,
            overlaps="association_artifacts_storages_rows,object_storage_row",
        )
    )

    def __str__(self) -> str:
        return (
            f"VFSStorageRow("
            f"id={self.id}, "
            f"name={self.name}, "
            f"host={self.host}, "
            f"base_path={self.base_path})"
        )

    def __repr__(self) -> str:
        return self.__str__()

    def to_dataclass(self) -> VFSStorageData:
        return VFSStorageData(
            id=self.id,
            name=self.name,
            host=self.host,
            base_path=Path(self.base_path),
        )



from __future__ import annotations

from collections.abc import AsyncIterator, Mapping
from contextlib import asynccontextmanager as actxmgr
from typing import Any
from urllib.parse import quote

import aiohttp

from ai.backend.common.dto.storage.request import (
    DeleteObjectReq,
    DownloadObjectReq,
    FileDeleteAsyncRequest,
    HuggingFaceGetCommitHashReqPathParam,
    HuggingFaceGetCommitHashReqQueryParam,
    HuggingFaceImportModelsReq,
    HuggingFaceRetrieveModelReqPathParam,
    HuggingFaceRetrieveModelReqQueryParam,
    HuggingFaceRetrieveModelsReq,
    HuggingFaceScanModelsReq,
    PresignedDownloadObjectReq,
    PresignedUploadObjectReq,
    ReservoirImportModelsReq,
    VFSDownloadFileReq,
    VFSListFilesReq,
)
from ai.backend.common.dto.storage.response import (
    FileDeleteAsyncResponse,
    HuggingFaceGetCommitHashResponse,
    HuggingFaceImportModelsResponse,
    HuggingFaceRetrieveModelResponse,
    HuggingFaceRetrieveModelsResponse,
    HuggingFaceScanModelsResponse,
    PresignedDownloadObjectResponse,
    PresignedUploadObjectResponse,
    ReservoirImportModelsResponse,
    VFolderCloneResponse,
    VFSListFilesResponse,
)
from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.clients.storage_proxy.base import StorageProxyHTTPClient
from ai.backend.manager.config.unified import StorageProxyClientTimeoutConfig
from ai.backend.manager.defs import DEFAULT_CHUNK_SIZE
from ai.backend.manager.errors.storage import UnexpectedStorageProxyResponseError

storage_proxy_client_resilience = Resilience(
    policies=[
        MetricPolicy(MetricArgs(domain=DomainType.CLIENT, layer=LayerType.STORAGE_PROXY_CLIENT)),
        RetryPolicy(
            RetryArgs(
                max_retries=3,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.EXPONENTIAL,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class StorageProxyManagerFacingClient:
    """
    Client for interacting with the storage proxy manager-facing API.
    This client provides methods for administrative operations such as volume management,
    folder creation/deletion, quota management, and performance metrics.
    """

    _client: StorageProxyHTTPClient
    _timeout_config: StorageProxyClientTimeoutConfig

    def __init__(
        self,
        client: StorageProxyHTTPClient,
        timeout_config: StorageProxyClientTimeoutConfig,
    ) -> None:
        self._client = client
        self._timeout_config = timeout_config

    @storage_proxy_client_resilience.apply()
    async def get_volumes(self) -> Mapping[str, Any]:
        """
        Get all volumes from the storage proxy.

        :return: Response containing volume information
        """
        return await self._client.request_with_response(
            "GET",
            "volumes",
            request_timeout=self._timeout_config.get_volumes.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def create_folder(
        self,
        volume: str,
        vfid: str,
        max_quota_scope_size: int | None = None,
        mode: int | None = None,
    ) -> None:
        """
        Create a new folder in the storage proxy.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param max_quota_scope_size: Maximum size for the quota scope (optional)
        :param mode: File mode permissions (optional)
        """
        options = {}
        if max_quota_scope_size and max_quota_scope_size > 0:
            options["initial_max_size_for_quota_scope"] = max_quota_scope_size
        body: dict[str, Any] = {
            "volume": volume,
            "vfid": vfid,
            "options": options,
        }
        if mode is not None:
            body["mode"] = mode
        await self._client.request(
            "POST",
            "folder/create",
            body=body,
            request_timeout=self._timeout_config.create_folder.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def delete_folder(
        self,
        volume: str,
        vfid: str,
    ) -> None:
        """
        Delete a folder from the storage proxy.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        """
        await self._client.request(
            "POST",
            "folder/delete",
            body={
                "volume": volume,
                "vfid": vfid,
            },
            request_timeout=self._timeout_config.delete_folder.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def clone_folder(
        self,
        src_volume: str,
        src_vfid: str,
        dst_volume: str,
        dst_vfid: str,
    ) -> VFolderCloneResponse:
        """
        Clone a folder to another location.

        :param src_volume: Source volume name
        :param src_vfid: Source virtual folder ID
        :param dst_volume: Destination volume name
        :param dst_vfid: Destination virtual folder ID
        """
        body = {
            "src_volume": src_volume,
            "src_vfid": src_vfid,
            "dst_volume": dst_volume,
            "dst_vfid": dst_vfid,
        }
        data = await self._client.request_with_response(
            "POST",
            "folder/clone",
            body=body,
            request_timeout=self._timeout_config.clone_folder.to_client_timeout(),
        )
        return VFolderCloneResponse.model_validate(data)

    @storage_proxy_client_resilience.apply()
    async def get_mount_path(
        self,
        volume: str,
        vfid: str,
        subpath: str = ".",
    ) -> Mapping[str, Any]:
        """
        Get the mount path for a folder.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param subpath: Subpath within the folder (default: ".")
        :return: Response containing the mount path
        """
        return await self._client.request_with_response(
            "GET",
            "folder/mount",
            body={
                "volume": volume,
                "vfid": vfid,
                "subpath": subpath,
            },
            request_timeout=self._timeout_config.get_mount_path.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def get_volume_hwinfo(
        self,
        volume: str,
    ) -> Mapping[str, Any]:
        """
        Get hardware information for a volume.

        :param volume: Volume name
        :return: Response containing hardware information
        """
        return await self._client.request_with_response(
            "GET",
            "volume/hwinfo",
            body={
                "volume": volume,
            },
            request_timeout=self._timeout_config.get_volume_hwinfo.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def get_volume_performance_metric(
        self,
        volume: str,
    ) -> Mapping[str, Any]:
        """
        Get performance metrics for a volume.

        :param volume: Volume name
        :return: Response containing performance metrics
        """
        return await self._client.request_with_response(
            "GET",
            "volume/performance-metric",
            body={
                "volume": volume,
            },
            request_timeout=self._timeout_config.get_volume_performance_metric.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def get_fs_usage(
        self,
        volume: str,
    ) -> Mapping[str, Any]:
        """
        Get filesystem usage information for a volume.

        :param volume: Volume name
        :return: Response containing filesystem usage
        """
        return await self._client.request_with_response(
            "GET",
            "folder/fs-usage",
            body={
                "volume": volume,
            },
            request_timeout=self._timeout_config.get_fs_usage.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def get_volume_quota(
        self,
        volume: str,
        vfid: str,
    ) -> Mapping[str, Any]:
        """
        Get quota information for a specific volume and virtual folder.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :return: Response containing quota information
        """
        return await self._client.request_with_response(
            "GET",
            "volume/quota",
            body={
                "volume": volume,
                "vfid": vfid,
            },
            request_timeout=self._timeout_config.get_volume_quota.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def update_volume_quota(
        self,
        volume: str,
        vfid: str,
        quota_scope_size: int,
    ) -> None:
        """
        Update the quota for a specific volume and virtual folder.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param quota_scope_size: New quota size to set
        """
        # DEPRECATED: This method is deprecated and will be removed in the future.
        body = {
            "volume": volume,
            "vfid": vfid,
            "size_bytes": quota_scope_size,
        }
        await self._client.request(
            "PATCH",
            "volume/quota",
            body=body,
            request_timeout=self._timeout_config.update_volume_quota.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def get_quota_scope(
        self,
        volume: str,
        qsid: str,
    ) -> Mapping[str, Any]:
        """
        Get quota scope information.

        :param volume: Volume name
        :param qsid: Quota scope ID
        :return: Response containing quota scope information
        """
        return await self._client.request_with_response(
            "GET",
            "quota-scope",
            body={
                "volume": volume,
                "qsid": qsid,
            },
            request_timeout=self._timeout_config.get_quota_scope.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def update_quota_scope(
        self,
        volume: str,
        qsid: str,
        max_vfolder_size: int,
    ) -> None:
        """
        Update quota scope settings.

        :param volume: Volume name
        :param qsid: Quota scope ID
        :param max_vfolder_size: Maximum size for the quota scope
        """
        body = {
            "volume": volume,
            "qsid": qsid,
            "options": {
                "limit_bytes": max_vfolder_size,
            },
        }
        await self._client.request(
            "PATCH",
            "quota-scope",
            body=body,
            request_timeout=self._timeout_config.update_quota_scope.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def delete_quota_scope_quota(
        self,
        volume: str,
        qsid: str,
    ) -> None:
        """
        Delete quota scope quota.

        :param volume: Volume name
        :param qsid: Quota scope ID
        """
        await self._client.request(
            "DELETE",
            "quota-scope/quota",
            body={
                "volume": volume,
                "qsid": qsid,
            },
            request_timeout=self._timeout_config.delete_quota_scope_quota.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def mkdir(
        self,
        *,
        volume: str,
        vfid: str,
        relpath: str | list[str],
        exist_ok: bool,
        parents: bool | None = None,
    ) -> Mapping[str, Any]:
        """
        Create a directory in a folder.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param relpath: Relative path of the directory to create
        :param exist_ok: If True, do not raise an error if the directory already exists
        :param parents: If True, create parent directories as needed
        :return: Response from the storage proxy
        """
        body = {
            "volume": volume,
            "vfid": vfid,
            "relpath": relpath,
            "exist_ok": exist_ok,
        }
        if parents is not None:
            body["parents"] = parents
        return await self._client.request_with_response(
            "POST",
            "folder/file/mkdir",
            body=body,
            request_timeout=self._timeout_config.mkdir.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def rename_file(
        self,
        volume: str,
        vfid: str,
        relpath: str,
        new_name: str,
    ) -> None:
        """
        Rename a file or directory.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param relpath: Current relative path of the file/directory
        :param new_name: New name for the file/directory
        """
        await self._client.request(
            "POST",
            "folder/file/rename",
            body={
                "volume": volume,
                "vfid": vfid,
                "relpath": relpath,
                "new_name": new_name,
            },
            request_timeout=self._timeout_config.rename_file.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def delete_files(
        self,
        volume: str,
        vfid: str,
        relpaths: list[str],
        recursive: bool = False,
    ) -> Mapping[str, Any]:
        """
        Delete files or directories.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param relpaths: List of relative paths to delete
        :param recursive: Whether to delete directories recursively
        :return: Response from the storage proxy
        """
        return await self._client.request_with_response(
            "POST",
            "folder/file/delete",
            body={
                "volume": volume,
                "vfid": vfid,
                "relpaths": relpaths,
                "recursive": recursive,
            },
            request_timeout=self._timeout_config.delete_files.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def delete_files_async(
        self,
        request: FileDeleteAsyncRequest,
    ) -> FileDeleteAsyncResponse:
        """
        Delete files or directories asynchronously using background tasks.

        :param request: Request containing volume, vfid, relpaths, and recursive flag
        :return: Response containing background task ID
        """
        response = await self._client.request_with_response(
            "POST",
            "folder/file/delete-async",
            body=request.model_dump(mode="json"),
            request_timeout=self._timeout_config.delete_files_async.to_client_timeout(),
        )
        return FileDeleteAsyncResponse.model_validate(response)

    @storage_proxy_client_resilience.apply()
    async def move_file(
        self,
        volume: str,
        vfid: str,
        src: str,
        dst: str,
    ) -> None:
        """
        Move a file or directory.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param src: Source relative path
        :param dst: Destination relative path
        """
        await self._client.request(
            "POST",
            "folder/file/move",
            body={
                "volume": volume,
                "vfid": vfid,
                "src_relpath": src,
                "dst_relpath": dst,
            },
            request_timeout=self._timeout_config.move_file.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def upload_file(
        self,
        volume: str,
        vfid: str,
        relpath: str,
        size: str,
    ) -> Mapping[str, Any]:
        """
        Upload a file to the storage proxy.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param relpath: Relative path of the file
        :param size: Size of the file
        :return: Response from the storage proxy
        """
        return await self._client.request_with_response(
            "POST",
            "folder/file/upload",
            body={
                "volume": volume,
                "vfid": vfid,
                "relpath": relpath,
                "size": size,
            },
            request_timeout=self._timeout_config.upload_file.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def download_file(
        self,
        *,
        volume: str,
        vfid: str,
        relpath: str,
        archive: bool = False,
        unmanaged_path: str | None = None,
    ) -> Mapping[str, Any]:
        """
        Download a file from the storage proxy.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param relpath: Relative path of the file
        :param archive: If True, download as an archive
        :param unmanaged_path: Optional unmanaged path for the file
        :return: Response from the storage proxy containing file data
        """
        return await self._client.request_with_response(
            "POST",
            "folder/file/download",
            body={
                "volume": volume,
                "vfid": vfid,
                "relpath": relpath,
                "archive": archive,
                "unmanaged_path": unmanaged_path,
            },
            request_timeout=self._timeout_config.download_file.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def list_files(
        self,
        volume: str,
        vfid: str,
        relpath: str,
    ) -> Mapping[str, Any]:
        """
        List files in a directory.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param relpath: Relative path of the directory
        :return: Response from the storage proxy containing file list
        """
        return await self._client.request_with_response(
            "POST",
            "folder/file/list",
            body={
                "volume": volume,
                "vfid": vfid,
                "relpath": relpath,
            },
            request_timeout=self._timeout_config.list_files.to_client_timeout(),
        )

    @actxmgr
    async def _fetch_file(
        self,
        volume: str,
        vfid: str,
        relpath: str,
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        """
        Fetch file content from the storage proxy.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param relpath: Relative path of the file
        :return: Response from the storage proxy containing file content
        """
        async with self._client.request_stream_response(
            "POST",
            "folder/file/fetch",
            body={
                "volume": volume,
                "vfid": vfid,
                "relpath": relpath,
            },
            request_timeout=self._timeout_config.fetch_file.to_client_timeout(),
        ) as response_stream:
            yield response_stream

    # TODO: There are some cases where `fetch_file_content` returns empty chunks are expected as successful scenarios.
    # Re-attach the commented-out decorator after refactoring the code.
    # @client_decorator()
    async def fetch_file_content(
        self,
        volume: str,
        vfid: str,
        relpath: str,
    ) -> bytes:
        """
        Fetch file content from the storage proxy.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param relpath: Relative path of the file
        :return: Response from the storage proxy containing file content
        """
        async with self._fetch_file(
            volume=volume,
            vfid=vfid,
            relpath=relpath,
        ) as response_stream:
            chunks = b""
            while True:
                chunk = await response_stream.content.read(DEFAULT_CHUNK_SIZE)
                if not chunk:
                    break
                chunks += chunk
            if not chunks:
                raise UnexpectedStorageProxyResponseError(
                    f"No content received for {volume}/{vfid}/{relpath}"
                )
            return chunks

    # TODO: Support AsyncIterator for `client_decorator`
    async def fetch_file_content_streaming(
        self,
        volume: str,
        vfid: str,
        relpath: str,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
    ) -> AsyncIterator[bytes]:
        """
        Fetch file content from the storage proxy as a byte-stream.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :param relpath: Relative path of the file
        :param chunk_size: Size of each chunk to read
        :yield: Chunks of bytes
        """
        async with self._fetch_file(volume=volume, vfid=vfid, relpath=relpath) as resp:
            resp.raise_for_status()

            async for chunk in resp.content.iter_chunked(chunk_size):
                if not chunk:
                    break
                yield chunk

    @storage_proxy_client_resilience.apply()
    async def get_folder_usage(
        self,
        volume: str,
        vfid: str,
    ) -> Mapping[str, Any]:
        """
        Get usage information for a virtual folder.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :return: Response from the storage proxy containing usage information
        """
        return await self._client.request_with_response(
            "GET",
            "folder/usage",
            body={
                "volume": volume,
                "vfid": vfid,
            },
            request_timeout=self._timeout_config.get_folder_usage.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def get_used_bytes(
        self,
        volume: str,
        vfid: str,
    ) -> Mapping[str, Any]:
        """
        Get the number of bytes used by a virtual folder.

        :param volume: Volume name
        :param vfid: Virtual folder ID
        :return: Response from the storage proxy containing used bytes
        """
        return await self._client.request_with_response(
            "GET",
            "folder/used-bytes",
            body={
                "volume": volume,
                "vfid": vfid,
            },
            request_timeout=self._timeout_config.get_used_bytes.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def scan_huggingface_models(
        self,
        req: HuggingFaceScanModelsReq,
    ) -> HuggingFaceScanModelsResponse:
        """
        Scan HuggingFace models in the specified registry.
        """
        resp = await self._client.request_with_response(
            "POST",
            "v1/registries/huggingface/scan",
            body=req.model_dump(by_alias=True),
            request_timeout=self._timeout_config.scan_huggingface_models.to_client_timeout(),
        )
        return HuggingFaceScanModelsResponse.model_validate(resp)

    @storage_proxy_client_resilience.apply()
    async def retrieve_huggingface_models(
        self,
        req: HuggingFaceRetrieveModelsReq,
    ) -> HuggingFaceRetrieveModelsResponse:
        """
        Retreive HuggingFace models in the specified registry.
        """
        resp = await self._client.request_with_response(
            "POST",
            "v1/registries/huggingface/models/batch",
            body=req.model_dump(by_alias=True),
            request_timeout=self._timeout_config.retrieve_huggingface_models.to_client_timeout(),
        )
        return HuggingFaceRetrieveModelsResponse.model_validate(resp)

    @storage_proxy_client_resilience.apply()
    async def retrieve_huggingface_model(
        self,
        path: HuggingFaceRetrieveModelReqPathParam,
        query: HuggingFaceRetrieveModelReqQueryParam,
    ) -> HuggingFaceRetrieveModelResponse:
        """
        Retreive HuggingFace single model in the specified registry.
        """
        encoded_model_id = quote(path.model_id, safe="")

        resp = await self._client.request_with_response(
            "GET",
            f"v1/registries/huggingface/model/{encoded_model_id}",
            params={
                "registry_name": query.registry_name,
                "revision": query.revision,
            },
            request_timeout=self._timeout_config.retrieve_huggingface_model.to_client_timeout(),
        )
        return HuggingFaceRetrieveModelResponse.model_validate(resp)

    @storage_proxy_client_resilience.apply()
    async def import_huggingface_models(
        self,
        req: HuggingFaceImportModelsReq,
    ) -> HuggingFaceImportModelsResponse:
        """
        Import multiple HuggingFace models into the specified registry.
        """
        resp = await self._client.request_with_response(
            "POST",
            "v1/registries/huggingface/import",
            body=req.model_dump(by_alias=True),
            request_timeout=self._timeout_config.import_huggingface_models.to_client_timeout(),
        )
        return HuggingFaceImportModelsResponse.model_validate(resp)

    @storage_proxy_client_resilience.apply()
    async def get_huggingface_model_commit_hash(
        self,
        path: HuggingFaceGetCommitHashReqPathParam,
        query: HuggingFaceGetCommitHashReqQueryParam,
    ) -> HuggingFaceGetCommitHashResponse:
        """
        Get the commit hash for a specific HuggingFace model revision.
        """
        params = {"registry_name": query.registry_name}
        if query.revision:
            params["revision"] = query.revision

        encoded_model_id = quote(path.model_id, safe="")
        resp = await self._client.request_with_response(
            "GET",
            f"v1/registries/huggingface/model/{encoded_model_id}/commit-hash",
            params=params,
            request_timeout=self._timeout_config.get_huggingface_model_commit_hash.to_client_timeout(),
        )

        return HuggingFaceGetCommitHashResponse.model_validate(resp)

    @storage_proxy_client_resilience.apply()
    async def import_reservoir_models(
        self,
        req: ReservoirImportModelsReq,
    ) -> ReservoirImportModelsResponse:
        """
        Import multiple Reservoir models into the specified registry.
        """
        resp = await self._client.request(
            "POST",
            "v1/registries/reservoir/import",
            body=req.model_dump(by_alias=True),
            request_timeout=self._timeout_config.import_reservoir_models.to_client_timeout(),
        )
        return ReservoirImportModelsResponse.model_validate(resp)

    @storage_proxy_client_resilience.apply()
    async def download_s3_file(
        self,
        storage_name: str,
        bucket_name: str,
        req: DownloadObjectReq,
    ) -> None:
        """
        Download a file from S3 storage.
        """
        await self._client.request_with_response(
            "POST",
            f"v1/storages/s3/{storage_name}/buckets/{bucket_name}/object/download",
            body=req.model_dump(by_alias=True),
            request_timeout=self._timeout_config.download_s3_file.to_client_timeout(),
        )

    @storage_proxy_client_resilience.apply()
    async def get_s3_presigned_download_url(
        self,
        storage_name: str,
        bucket_name: str,
        req: PresignedDownloadObjectReq,
    ) -> PresignedDownloadObjectResponse:
        """
        Get a presigned URL for downloading an object from storage.
        """
        resp = await self._client.request_with_response(
            "POST",
            f"v1/storages/s3/{storage_name}/buckets/{bucket_name}/object/presigned/download",
            body=req.model_dump(by_alias=True),
            request_timeout=self._timeout_config.get_s3_presigned_download_url.to_client_timeout(),
        )
        return PresignedDownloadObjectResponse.model_validate(resp)

    @storage_proxy_client_resilience.apply()
    async def get_s3_presigned_upload_url(
        self,
        storage_name: str,
        bucket_name: str,
        req: PresignedUploadObjectReq,
    ) -> PresignedUploadObjectResponse:
        """
        Get a presigned URL for uploading an object to storage.
        """
        resp = await self._client.request_with_response(
            "POST",
            f"v1/storages/s3/{storage_name}/buckets/{bucket_name}/object/presigned/upload",
            body=req.model_dump(by_alias=True),
            request_timeout=self._timeout_config.get_s3_presigned_upload_url.to_client_timeout(),
        )
        return PresignedUploadObjectResponse.model_validate(resp)

    @storage_proxy_client_resilience.apply()
    async def delete_s3_object(
        self,
        storage_name: str,
        bucket_name: str,
        req: DeleteObjectReq,
    ) -> None:
        """
        Delete a file from S3 storage.
        """
        await self._client.request(
            "DELETE",
            f"v1/storages/s3/{storage_name}/buckets/{bucket_name}/object",
            body=req.model_dump(by_alias=True),
            request_timeout=self._timeout_config.delete_s3_object.to_client_timeout(),
        )

    # TODO: Support storage_proxy_client_resilience
    @actxmgr
    async def download_vfs_file_streaming(
        self,
        storage_name: str,
        req: VFSDownloadFileReq,
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        """
        Download a file from VFS storage using streaming.

        :param storage_name: Name of the VFS storage
        :param req: VFS download file request
        :return: Streaming response from the storage proxy
        """
        async with self._client.request_stream_response(
            "POST",
            f"v1/storages/vfs/{storage_name}/download",
            body=req.model_dump(by_alias=True),
            request_timeout=self._timeout_config.download_vfs_file_streaming.to_client_timeout(),
        ) as response_stream:
            yield response_stream

    @storage_proxy_client_resilience.apply()
    async def list_vfs_files(
        self,
        storage_name: str,
        req: VFSListFilesReq,
    ) -> VFSListFilesResponse:
        """
        List files recursively in a VFS storage directory.

        :param storage_name: Name of the VFS storage
        :param req: VFS list files request
        :return: Response containing list of files with metadata
        """
        resp = await self._client.request_with_response(
            "GET",
            f"v1/storages/vfs/{storage_name}/files",
            body=req.model_dump(by_alias=True),
            request_timeout=self._timeout_config.list_vfs_files.to_client_timeout(),
        )
        return VFSListFilesResponse.model_validate(resp)



"""Session and kernel lifecycle data types."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from typing import Any
from uuid import UUID

from ai.backend.common.types import (
    AccessKey,
    AgentId,
    BinarySize,
    ClusterMode,
    KernelId,
    ResourceSlot,
    SessionId,
    SessionTypes,
)
from ai.backend.manager.data.kernel.types import KernelInfo
from ai.backend.manager.data.session.types import SessionInfo
from ai.backend.manager.defs import DEFAULT_ROLE
from ai.backend.manager.errors.kernel import MainKernelNotFound, TooManyKernelsFound
from ai.backend.manager.models.kernel import KernelStatus
from ai.backend.manager.models.network import NetworkType

from .image import ImageConfigData


@dataclass
class KernelBindingData:
    """Kernel-agent binding data for precondition checking and session starting."""

    kernel_id: KernelId
    agent_id: AgentId | None
    agent_addr: str | None
    scaling_group: str
    image: str
    architecture: str
    status: KernelStatus | None = None
    status_changed: float | None = None
    cluster_role: str = DEFAULT_ROLE
    cluster_idx: int = 0
    local_rank: int = 0
    cluster_hostname: str | None = None
    uid: int | None = None
    main_gid: int | None = None
    gids: list[int] = field(default_factory=list)
    requested_slots: ResourceSlot = field(default_factory=ResourceSlot)
    resource_opts: dict[str, Any] = field(default_factory=dict)
    bootstrap_script: str | None = None
    startup_command: str | None = None
    preopen_ports: list[int] = field(default_factory=list)
    internal_data: dict[str, Any] | None = None
    vfolder_mounts: list[Any] = field(
        default_factory=list
    )  # Would be list[VFolderMount] in full impl


@dataclass
class SessionDataForPull:
    """Data for a session that needs image pulling."""

    session_id: SessionId
    creation_id: str
    access_key: AccessKey
    kernels: list[KernelBindingData]


@dataclass
class SessionDataForStart:
    """Data for a session ready to start with full details."""

    session_id: SessionId
    creation_id: str
    access_key: AccessKey
    session_type: SessionTypes
    name: str
    cluster_mode: ClusterMode
    kernels: list[KernelBindingData]
    user_uuid: UUID
    user_email: str
    user_name: str
    environ: dict[str, str]
    network_type: NetworkType | None = None
    network_id: str | None = None


@dataclass
class ScheduledSessionData:
    """Data for a scheduled session ready for precondition check."""

    session_id: SessionId
    creation_id: str
    access_key: AccessKey
    session_type: SessionTypes
    name: str
    kernels: list[KernelBindingData]
    # Additional fields for PREPARED sessions
    cluster_mode: ClusterMode | None = None
    user_uuid: UUID | None = None
    user_email: str | None = None
    user_name: str | None = None
    network_type: NetworkType | None = None
    network_id: str | None = None


@dataclass
class SessionsForPullWithImages:
    """Sessions for image pulling with their image configurations."""

    sessions: list[SessionDataForPull]
    image_configs: dict[str, ImageConfigData]


@dataclass
class SessionsForStartWithImages:
    """Sessions for starting with their image configurations."""

    sessions: list[SessionDataForStart]
    image_configs: dict[str, ImageConfigData]


@dataclass
class ScheduledSessionsWithImages:
    """Scheduled sessions with their image configurations."""

    sessions: list[ScheduledSessionData]
    image_configs: dict[str, ImageConfigData]


@dataclass
class KernelStartData:
    """Kernel data for starting a session."""

    kernel_id: UUID
    agent_id: AgentId
    agent_addr: str
    scaling_group: str
    image: str
    architecture: str
    cluster_role: str
    cluster_idx: int
    requested_slots: ResourceSlot
    resource_opts: dict[str, Any]
    preopen_ports: list[int]
    container_id: str | None = None
    cluster_hostname: str | None = None
    bootstrap_script: str | None = None
    startup_command: str | None = None


@dataclass
class PreparedSessionData:
    """Data for a prepared session ready to start."""

    session_id: SessionId
    creation_id: str
    access_key: AccessKey
    session_type: SessionTypes
    name: str
    cluster_mode: ClusterMode
    kernels: list[KernelStartData]
    user_uuid: UUID
    user_email: str
    user_name: str
    network_type: NetworkType | None = None
    network_id: str | None = None


@dataclass
class PreparedSessionsWithImages:
    """Prepared sessions with their image configurations."""

    sessions: list[PreparedSessionData]
    image_configs: dict[str, ImageConfigData]


@dataclass
class KernelCreationInfo:
    """Information about kernel creation from agent."""

    container_id: str | None = None
    resource_spec: dict[str, Any] | None = None
    attached_devices: dict[str, Any] = field(default_factory=dict)
    repl_in_port: int | None = None
    repl_out_port: int | None = None
    stdin_port: int | None = None
    stdout_port: int | None = None
    service_ports: list[int] = field(default_factory=list)
    kernel_host: str | None = None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> KernelCreationInfo:
        """Create from dictionary, handling missing or invalid fields."""
        return cls(
            container_id=data.get("container_id"),
            resource_spec=data.get("resource_spec"),
            attached_devices=data.get("attached_devices", {}),
            repl_in_port=data.get("repl_in_port"),
            repl_out_port=data.get("repl_out_port"),
            stdin_port=data.get("stdin_port"),
            stdout_port=data.get("stdout_port"),
            service_ports=data.get("service_ports", []),
            kernel_host=data.get("kernel_host"),
        )

    def get_resource_allocations(self) -> ResourceSlot:
        """
        Extract resource allocations from resource_spec.
        Compatible with AgentRegistry.convert_resource_spec_to_resource_slot() format.

        Handles the agent-side nested format:
        allocations: {
            "device_type": {
                "slot_name": {
                    "device_id": "value"
                }
            }
        }
        """
        if not self.resource_spec or "allocations" not in self.resource_spec:
            return ResourceSlot()

        allocations = self.resource_spec["allocations"]
        return self.convert_allocations_to_resource_slot(allocations)

    @staticmethod
    def convert_allocations_to_resource_slot(allocations: dict[str, Any]) -> ResourceSlot:
        """
        Convert per-device resource spec allocations (agent-side format)
        back into a resource slot (manager-side format).

        This is a static method that mirrors AgentRegistry.convert_resource_spec_to_resource_slot()
        for compatibility.

        Args:
            allocations: The allocations dict from resource_spec

        Returns:
            ResourceSlot with aggregated resource values
        """
        if not allocations or not isinstance(allocations, dict):
            return ResourceSlot()

        slots = ResourceSlot()

        # Handle the nested structure from agent
        for alloc_map in allocations.values():
            if not isinstance(alloc_map, dict):
                continue

            for slot_name, allocation_by_device in alloc_map.items():
                if not isinstance(allocation_by_device, dict):
                    # If it's not the expected nested structure,
                    # try to use it directly as a value
                    if allocation_by_device is not None:
                        slots[slot_name] = str(allocation_by_device)
                    continue

                # Sum allocations across devices
                total_allocs: list[Decimal] = []
                for allocation in allocation_by_device.values():
                    if allocation is None:
                        continue

                    # Handle BinarySize values (e.g., "1073741824b", "1g")
                    if (
                        isinstance(allocation, str)
                        and len(allocation) > 0
                        and BinarySize.suffix_map.get(allocation[-1].lower()) is not None
                    ):
                        total_allocs.append(Decimal(BinarySize.from_str(allocation)))
                    else:
                        # Regular decimal value or special values like "Infinity"
                        total_allocs.append(Decimal(allocation))

                if total_allocs:
                    slots[slot_name] = str(sum(total_allocs))

        return slots


@dataclass(frozen=True)
class SessionRunningData:
    """
    Data for updating a session to RUNNING state.
    Contains the calculated occupying_slots from all kernels.
    """

    session_id: SessionId
    occupying_slots: ResourceSlot


@dataclass
class SessionWithKernels:
    """
    Bundles a session with its associated kernels.

    This is the primary data unit for scheduler operations,
    representing a session and all its kernels as an atomic unit.

    Attributes:
        session_info: Session information including lifecycle data
        kernel_infos: List of kernels belonging to this session
        phase_attempts: Number of attempts for current phase from scheduling history
                       (used for failure classification: give_up when >= max_retries)
        phase_started_at: When the current phase started from scheduling history
                         (used for failure classification: expired when timeout exceeded)
    """

    session_info: SessionInfo
    kernel_infos: list[KernelInfo]
    phase_attempts: int = 0
    phase_started_at: datetime | None = None

    @property
    def main_kernel(self) -> KernelInfo:
        """Get the main kernel (kernel with DEFAULT_ROLE as cluster_role)."""
        main_kernels = [k for k in self.kernel_infos if k.cluster.cluster_role == DEFAULT_ROLE]
        if len(main_kernels) > 1:
            raise TooManyKernelsFound(
                f"Session {self.session_info.identity.id} has more than 1 main kernel"
            )
        if len(main_kernels) == 0:
            raise MainKernelNotFound(f"Session {self.session_info.identity.id} has no main kernel")
        return main_kernels[0]



"""add deploying lifecycle state and revision_history_limit column

Revision ID: 5e7a3b9c1d2f
Revises: 4c9e2f3a5b6d
Create Date: 2025-12-17

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "5e7a3b9c1d2f"
down_revision = "4c9e2f3a5b6d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Add 'deploying' value to endpointlifecycle enum
    op.execute("ALTER TYPE endpointlifecycle ADD VALUE IF NOT EXISTS 'deploying'")

    # Add revision_history_limit column to endpoints table
    op.add_column(
        "endpoints",
        sa.Column(
            "revision_history_limit",
            sa.Integer(),
            nullable=False,
            server_default=sa.text("10"),
        ),
    )


def downgrade() -> None:
    # Drop revision_history_limit column
    op.drop_column("endpoints", "revision_history_limit")
    # Note: PostgreSQL doesn't support removing enum values directly
    # The 'deploying' value will remain in the enum type



"""migrate user data to roles

Revision ID: a4d56e86d9ee
Revises: 9adcd6f48ba1
Create Date: 2025-08-06 21:28:29.354670

"""

import enum
import uuid
from collections.abc import Iterable, Mapping, Sequence
from typing import Any

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection, Row

from ai.backend.manager.models.base import EnumValueType, IDColumn
from ai.backend.manager.models.rbac_models.migration.enums import (
    ScopeType,
)
from ai.backend.manager.models.rbac_models.migration.models import (
    get_permission_groups_table,
    get_permissions_table,
    get_roles_table,
    get_user_roles_table,
    mapper_registry,
)
from ai.backend.manager.models.rbac_models.migration.types import (
    PermissionCreateInput,
    PermissionGroupCreateInput,
    UserRoleCreateInput,
)
from ai.backend.manager.models.rbac_models.migration.user import (
    ENTITY_TYPES_IN_ROLE,
    OPERATIONS_IN_ROLE,
    UserData,
    get_user_self_role_creation_input,
)
from ai.backend.manager.models.rbac_models.migration.utils import (
    insert_if_data_exists,
    query_permission_groups_by_scope_ids,
    query_role_rows_by_name,
)

# revision identifiers, used by Alembic.
revision = "a4d56e86d9ee"
down_revision = "9adcd6f48ba1"
branch_labels = None
depends_on = None


class UserRole(enum.StrEnum):
    """
    User's role.
    """

    SUPERADMIN = "superadmin"
    ADMIN = "admin"
    USER = "user"
    MONITOR = "monitor"


class Tables:
    @staticmethod
    def get_users_table() -> sa.Table:
        return sa.Table(
            "users",
            mapper_registry.metadata,
            IDColumn("uuid"),
            sa.Column("username", sa.String(length=64), unique=True),
            sa.Column("domain_name", sa.String(length=64), index=True),
            sa.Column("role", EnumValueType(UserRole), default=UserRole.USER),
            extend_existing=True,
        )


class RoleCreator:
    @classmethod
    def _create_self_roles(
        cls, db_conn: Connection, rows: Sequence[Row[Any]]
    ) -> dict[str, uuid.UUID]:
        roles_table = get_roles_table()
        role_inputs: list[dict[str, Any]] = []
        role_name_user_id_map: dict[str, uuid.UUID] = {}
        for row in rows:
            data = UserData.from_row(row)
            role_input = get_user_self_role_creation_input(data)
            role_inputs.append(role_input.to_dict())
            role_name_user_id_map[role_input.name] = data.id
        insert_if_data_exists(db_conn, roles_table, role_inputs)
        return role_name_user_id_map

    @classmethod
    def _create_permission_groups(
        cls, db_conn: Connection, role_id_user_id_map: Mapping[uuid.UUID, uuid.UUID]
    ) -> None:
        permission_groups_table = get_permission_groups_table()
        permission_group_inputs: list[dict[str, Any]] = []
        for role_id, user_id in role_id_user_id_map.items():
            input = PermissionGroupCreateInput(
                role_id=role_id,
                scope_type=ScopeType.USER,
                scope_id=str(user_id),
            )
            permission_group_inputs.append(input.to_dict())
        insert_if_data_exists(db_conn, permission_groups_table, permission_group_inputs)

    @classmethod
    def _create_permissions(
        cls, db_conn: Connection, permission_group_ids: Iterable[uuid.UUID]
    ) -> None:
        permission_inputs: list[dict[str, Any]] = []
        for permission_group_id in permission_group_ids:
            for entity_type in ENTITY_TYPES_IN_ROLE:
                for operation in OPERATIONS_IN_ROLE:
                    input = PermissionCreateInput(
                        permission_group_id=permission_group_id,
                        entity_type=entity_type,
                        operation=operation,
                    )
                    permission_inputs.append(input.to_dict())
        insert_if_data_exists(db_conn, get_permissions_table(), permission_inputs)

    @classmethod
    def _query_user_row(
        cls, db_conn: Connection, offset: int, page_size: int
    ) -> Sequence[Row[Any]]:
        """
        Query all user rows with pagination.
        """
        users_table = Tables.get_users_table()
        user_query = (
            sa.select(
                users_table.c.uuid,
                users_table.c.username,
                users_table.c.domain_name,
                users_table.c.role,
            )
            .offset(offset)
            .limit(page_size)
            .order_by(users_table.c.uuid)
        )
        return db_conn.execute(user_query).all()

    @classmethod
    def create_user_self_roles_and_permissions(cls, db_conn: Connection) -> None:
        """
        Migrate user data to roles and permissions.
        All users have a default self role and permissions.
        """
        offset = 0
        page_size = 1000
        while True:
            rows = cls._query_user_row(db_conn, offset, page_size)
            offset += page_size
            if not rows:
                break
            role_name_user_id_map = cls._create_self_roles(db_conn, rows)
            role_rows = query_role_rows_by_name(db_conn, list(role_name_user_id_map.keys()))
            role_id_user_id_map: dict[uuid.UUID, uuid.UUID] = {
                row.id: role_name_user_id_map[row.name] for row in role_rows
            }
            cls._create_permission_groups(db_conn, role_id_user_id_map)
            str_user_ids = [str(user_id) for user_id in role_id_user_id_map.values()]
            permission_group_ids = query_permission_groups_by_scope_ids(db_conn, str_user_ids)
            cls._create_permissions(db_conn, permission_group_ids)


class RoleMapper:
    @classmethod
    def _create_user_roles(
        cls, db_conn: Connection, role_id_user_id_map: Mapping[uuid.UUID, uuid.UUID]
    ) -> None:
        user_roles_table = get_user_roles_table()
        user_role_inputs: list[dict[str, Any]] = []
        for role_id, user_id in role_id_user_id_map.items():
            user_role_input = UserRoleCreateInput(user_id=user_id, role_id=role_id)
            user_role_inputs.append(user_role_input.to_dict())
        insert_if_data_exists(db_conn, user_roles_table, user_role_inputs)

    @classmethod
    def _query_user_row(
        cls, db_conn: Connection, offset: int, page_size: int
    ) -> Sequence[Row[Any]]:
        """
        Query all user rows with pagination.
        """
        users_table = Tables.get_users_table()
        user_query = (
            sa.select(
                users_table.c.uuid,
                users_table.c.username,
                users_table.c.domain_name,
                users_table.c.role,
            )
            .offset(offset)
            .limit(page_size)
            .order_by(users_table.c.uuid)
        )
        return db_conn.execute(user_query).all()

    @classmethod
    def map_users_to_roles(cls, db_conn: Connection) -> None:
        """
        Map users to their self roles.
        """
        offset = 0
        page_size = 1000
        while True:
            rows = cls._query_user_row(db_conn, offset, page_size)
            offset += page_size
            if not rows:
                break
            role_name_user_id_map: dict[str, uuid.UUID] = {}
            for row in rows:
                data = UserData.from_row(row)
                role_input = get_user_self_role_creation_input(data)
                role_name_user_id_map[role_input.name] = data.id

            role_rows = query_role_rows_by_name(db_conn, list(role_name_user_id_map.keys()))
            role_id_user_id_map: dict[uuid.UUID, uuid.UUID] = {
                row.id: role_name_user_id_map[row.name] for row in role_rows
            }
            cls._create_user_roles(db_conn, role_id_user_id_map)


def upgrade() -> None:
    conn = op.get_bind()
    RoleCreator.create_user_self_roles_and_permissions(conn)
    RoleMapper.map_users_to_roles(conn)


def downgrade() -> None:
    pass



"""Type definitions for resource preset repository."""

from __future__ import annotations

from dataclasses import dataclass

from ai.backend.common.json import dump_json, load_json
from ai.backend.common.types import ResourceSlot

from .db_source.types import PerScalingGroupResourceData, PresetAllocatabilityData


@dataclass
class CheckPresetsResult:
    """Result of checking resource presets from repository."""

    presets: list[PresetAllocatabilityData]
    keypair_limits: ResourceSlot
    keypair_using: ResourceSlot
    keypair_remaining: ResourceSlot
    group_limits: ResourceSlot
    group_using: ResourceSlot
    group_remaining: ResourceSlot
    scaling_group_remaining: ResourceSlot
    scaling_groups: dict[str, PerScalingGroupResourceData]

    def to_cache(self) -> bytes:
        """
        Serialize the result to a JSON string for caching.

        :return: JSON bytes representation of the result
        """
        cache_data = {
            "presets": [p.to_cache() for p in self.presets],
            "keypair_limits": self.keypair_limits.to_json(),
            "keypair_using": self.keypair_using.to_json(),
            "keypair_remaining": self.keypair_remaining.to_json(),
            "group_limits": self.group_limits.to_json(),
            "group_using": self.group_using.to_json(),
            "group_remaining": self.group_remaining.to_json(),
            "scaling_group_remaining": self.scaling_group_remaining.to_json(),
            "scaling_groups": {
                sgname: sg_data.to_cache() for sgname, sg_data in self.scaling_groups.items()
            },
        }

        return dump_json(cache_data)

    @classmethod
    def from_cache(cls, cached_data: bytes) -> CheckPresetsResult:
        """
        Deserialize the result from a cached JSON string.

        :param cached_data: JSON bytes from cache
        :return: CheckPresetsResult instance
        """
        data = load_json(cached_data)

        return cls(
            presets=[PresetAllocatabilityData.from_cache(p) for p in data["presets"]],
            keypair_limits=ResourceSlot.from_json(data["keypair_limits"]),
            keypair_using=ResourceSlot.from_json(data["keypair_using"]),
            keypair_remaining=ResourceSlot.from_json(data["keypair_remaining"]),
            group_limits=ResourceSlot.from_json(data["group_limits"]),
            group_using=ResourceSlot.from_json(data["group_using"]),
            group_remaining=ResourceSlot.from_json(data["group_remaining"]),
            scaling_group_remaining=ResourceSlot.from_json(data["scaling_group_remaining"]),
            scaling_groups={
                sgname: PerScalingGroupResourceData.from_cache(sg_data)
                for sgname, sg_data in data["scaling_groups"].items()
            },
        )



"""Repository for fetching container registry information for quota management."""

from __future__ import annotations

from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.container_registry.types import PerProjectContainerRegistryInfo
from ai.backend.manager.models.rbac import ProjectScope
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.container_registry_quota.db_source import (
    PerProjectRegistryQuotaDBSource,
)

per_project_registry_quota_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(
                domain=DomainType.REPOSITORY,
                layer=LayerType.CONTAINER_REGISTRY_REPOSITORY,
            )
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


# NOTE: Only one implementation exists for now, so no ABC is used.
# Introduce an abstract base class when multiple implementations are needed.
class PerProjectRegistryQuotaRepository:
    _db_source: PerProjectRegistryQuotaDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = PerProjectRegistryQuotaDBSource(db)

    @per_project_registry_quota_repository_resilience.apply()
    async def fetch_container_registry_row(
        self, scope_id: ProjectScope
    ) -> PerProjectContainerRegistryInfo:
        return await self._db_source.fetch_container_registry_row(scope_id)



"""Type definitions for resource preset DB source."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any
from uuid import UUID

from ai.backend.common.types import (
    AccessKey,
    BinarySize,
    ResourceSlot,
    SlotName,
    SlotTypes,
)
from ai.backend.common.types import (
    LegacyResourceSlotState as ResourceSlotState,
)
from ai.backend.manager.data.resource_preset.types import ResourcePresetData
from ai.backend.manager.models.kernel import KernelRow


@dataclass
class ResourceUsageData:
    """Resource usage data with limits, occupied, and remaining."""

    limits: ResourceSlot
    occupied: ResourceSlot
    remaining: ResourceSlot


@dataclass
class KeypairResourceData:
    """Resource data for a keypair."""

    limits: ResourceSlot
    occupied: ResourceSlot
    remaining: ResourceSlot
    group_limits: ResourceSlot
    group_occupied: ResourceSlot
    group_remaining: ResourceSlot
    scaling_group_remaining: ResourceSlot


@dataclass
class PerScalingGroupResourceData:
    """Resource data per scaling group."""

    using: ResourceSlot
    remaining: ResourceSlot

    def to_cache(self) -> dict[str, Any]:
        """Serialize to cache-friendly format."""
        return {
            str(ResourceSlotState.OCCUPIED): self.using.to_json(),
            str(ResourceSlotState.AVAILABLE): self.remaining.to_json(),
        }

    @classmethod
    def from_cache(cls, data: dict[str, Any]) -> PerScalingGroupResourceData:
        """Deserialize from cache format."""
        return cls(
            using=ResourceSlot.from_json(data[ResourceSlotState.OCCUPIED]),
            remaining=ResourceSlot.from_json(data[ResourceSlotState.AVAILABLE]),
        )


@dataclass
class PresetAllocatabilityData:
    """Preset with allocatability information."""

    preset: ResourcePresetData
    allocatable: bool

    def to_cache(self) -> dict[str, Any]:
        """Serialize to cache-friendly format."""
        return {
            "preset": {
                "id": str(self.preset.id),
                "name": self.preset.name,
                "resource_slots": self.preset.resource_slots.to_json(),
                "shared_memory": str(self.preset.shared_memory)
                if self.preset.shared_memory is not None
                else None,
                "scaling_group_name": self.preset.scaling_group_name,
            },
            "allocatable": self.allocatable,
        }

    @classmethod
    def from_cache(cls, data: dict[str, Any]) -> PresetAllocatabilityData:
        """Deserialize from cache format."""
        return cls(
            preset=ResourcePresetData(
                id=UUID(data["preset"]["id"]),
                name=data["preset"]["name"],
                resource_slots=ResourceSlot.from_json(data["preset"]["resource_slots"]),
                shared_memory=int(BinarySize.from_str(data["preset"]["shared_memory"]))
                if data["preset"]["shared_memory"] is not None
                else None,
                scaling_group_name=data["preset"]["scaling_group_name"],
            ),
            allocatable=data["allocatable"],
        )


@dataclass
class CheckPresetsDBData:
    """All data fetched from DB for checking presets."""

    known_slot_types: Mapping[SlotName, SlotTypes]
    keypair_data: KeypairResourceData
    per_sgroup_data: dict[str, PerScalingGroupResourceData]
    presets: list[PresetAllocatabilityData]


class ResourceOccupancyFilter(ABC):
    """Abstract base class for resource occupancy filters."""

    @abstractmethod
    def get_condition(self) -> Any:
        """Return SQLAlchemy condition for this filter."""
        pass


class AccessKeyFilter(ResourceOccupancyFilter):
    """Filter for access key."""

    def __init__(self, access_key: AccessKey) -> None:
        self.access_key = access_key

    def get_condition(self) -> Any:
        return KernelRow.access_key == self.access_key


class GroupIdFilter(ResourceOccupancyFilter):
    """Filter for group ID."""

    def __init__(self, group_id: UUID) -> None:
        self.group_id = group_id

    def get_condition(self) -> Any:
        return KernelRow.group_id == self.group_id


class DomainNameFilter(ResourceOccupancyFilter):
    """Filter for domain name."""

    def __init__(self, domain_name: str) -> None:
        self.domain_name = domain_name

    def get_condition(self) -> Any:
        return KernelRow.domain_name == self.domain_name



from __future__ import annotations

from collections.abc import Awaitable, Callable, Iterable, Mapping
from contextlib import AbstractAsyncContextManager
from typing import (
    TYPE_CHECKING,
)

import aiohttp_cors
from aiohttp import web
from aiohttp.typedefs import Middleware

if TYPE_CHECKING:
    from .context import RootContext


type WebRequestHandler = Callable[
    [web.Request],
    Awaitable[web.StreamResponse],
]
type WebMiddleware = Middleware

type CORSOptions = Mapping[str, aiohttp_cors.ResourceOptions]
type AppCreator = Callable[
    [CORSOptions],
    tuple[web.Application, Iterable[WebMiddleware]],
]

type CleanupContext = Callable[["RootContext"], AbstractAsyncContextManager[None]]



from dataclasses import dataclass, field
from typing import Any, override

from ai.backend.manager.types import PartialModifier, TriState


@dataclass
class ArtifactModifier(PartialModifier):
    readonly: TriState[bool] = field(default_factory=TriState[bool].nop)
    description: TriState[str] = field(default_factory=TriState[str].nop)

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.readonly.update_dict(to_update, "readonly")
        self.description.update_dict(to_update, "description")
        return to_update
